
import numpy as np
import pyvista as pv
import matplotlib.pyplot as plt
import sys
from matplotlib import rc


def add_arrow(line, position=None, direction='right', size=15, color=None):
    """
    add an arrow to a line.

    line:       Line2D object
    position:   x-position of the arrow. If None, mean of xdata is taken
    direction:  'left' or 'right'
    size:       size of the arrow in fontsize points
    color:      if None, line color is taken.
    """
    if color is None:
        color = line.get_color()

    xdata = line.get_xdata()
    ydata = line.get_ydata()

    if position is None:
        position = xdata.mean()
    # find closest index
    start_ind = 0 #np.argmin(np.absolute(xdata - position))
    if direction == 'right':
        end_ind = start_ind + 1
    else:
        end_ind = start_ind - 1

    Px = xdata[end_ind] - (xdata[end_ind] - xdata[start_ind]) / 2.0
    Py = ydata[end_ind] - (ydata[end_ind] - ydata[start_ind]) / 2.0

    PPx = xdata[start_ind] + (xdata[end_ind] - xdata[start_ind]) / 2.01
    PPy = ydata[start_ind] + (ydata[end_ind] - ydata[start_ind]) / 2.01


    line.axes.annotate('',
        xytext=(PPx,PPy),

        xy=(Px, Py),
        arrowprops=dict(arrowstyle="->", color=color),
        size=size
    )


def wavelength_to_rgb(wavelength, gamma=1.0):
    """wavelength_to_rgb.

    Parameters
    ----------
    wavelength :
        wavelength
    gamma :
        gamma
    """
    wavelength = float(wavelength)
    if (380 <= wavelength <= 440):
        attenuation = (0.3 + ((0.7 * (wavelength - 380)) / (440 - 380)))
        R = ((((- (wavelength - 440)) / (440 - 380)) * attenuation) ** gamma)
        G = 0.0
        B = ((1.0 * attenuation) ** gamma)
    elif (440 <= wavelength <= 490):
        R = 0.0
        G = (((wavelength - 440) / (490 - 440)) ** gamma)
        B = 1.0
    elif (490 <= wavelength <= 510):
        R = 0.0
        G = 1.0
        B = (((- (wavelength - 510)) / (510 - 490)) ** gamma)
    elif (510 <= wavelength <= 580):
        R = (((wavelength - 510) / (580 - 510)) ** gamma)
        G = 1.0
        B = 0.0
    elif (580 <= wavelength <= 645):
        R = 1.0
        G = (((- (wavelength - 645)) / (645 - 580)) ** gamma)
        B = 0.0
    elif (645 <= wavelength <= 750):
        attenuation = (0.3 + ((0.7 * (750 - wavelength)) / (750 - 645)))
        R = ((1.0 * attenuation) ** gamma)
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0
    R *= 255
    G *= 255
    B *= 255
    return [(int(R) / 255.0), (int(G) / 255.0), (int(B) / 255.0)]

def display3d(SYSTEM, RAYS, view=0, inline=False,     BackgCol= 'white', BackgColTop = 'white', GridCol="black"):


    """display3d.

    Parameters
    ----------
    SYSTEM :
        SYSTEM
    RAYS :
        RAYS
    view :
        view
    """

    ST1="KrakenOS v0.1. Executing Script: " + sys.argv[0]
    OPA = 0.95

    CCC = pv.MultiBlock()
    INST = isinstance(RAYS, list)
    if INST == True:
        for R in RAYS:
            for rays in R.CC:
                RAY_VTK_OBJ = pv.lines_from_points(rays)
                CCC.append(RAY_VTK_OBJ)
    else:
        for rays in RAYS.CC:
            RAY_VTK_OBJ = pv.lines_from_points(rays)
            CCC.append(RAY_VTK_OBJ)


    p = pv.Plotter(shape=(1, 1), title=ST1,notebook=inline)
    Absorb_color = np.array([(10 / 256), (23 / 256), (24 / 256)])
    Mirror_color = np.array([(189 / 256), (189 / 256), (189 / 256)])
    # Glass_color = np.array([(190 / 256), (238 / 256), (246 / 256)])
    # Glass_color = np.array([(15 / 256), (141 / 256), (147 / 256)])
    Glass_color=np.array([12/256, 238/256, 246/256])
    recorte = view
    NN = SYSTEM.AAA.n_blocks
    if (SYSTEM.SDT[0].Drawing == 0):
        points2 = np.c_[(0, 0, 0)]
        c = pv.PolyData(points2)
        cc = pv.PolyData(points2)
    if (SYSTEM.SDT[0].Drawing == 1):
        c = SYSTEM.AAA[0]
        cc = SYSTEM.AAA[0]

    for n in range(1, NN):
        if (SYSTEM.SDT[n].Drawing == 1):
            AAAA = SYSTEM.AAA[n]
            if (SYSTEM.SDT[n].Glass != 'NULL'):
                if (SYSTEM.SDT[n].Color == [0, 0, 0]):
                    if (SYSTEM.SDT[n].Glass == 'MIRROR'):
                        color = Mirror_color
                    else:
                        color = Glass_color
                    if (SYSTEM.SDT[n].Glass == 'ABSORB'):
                        color = Absorb_color
                else:
                    color = SYSTEM.SDT[n].Color

                cc = cc.merge(AAAA)
                if (recorte == 1):
                    clippedx = AAAA.clip((1, 0, 0), invert=False)
                    clippedy = clippedx.clip((0, 1, 0), invert=False)
                    c = c.merge(clippedy)
                    clippedx = AAAA.clip(((- 1), 0, 0), invert=False)
                    clippedy = clippedx.clip((0, (- 1), 0), invert=False)
                    c = c.merge(clippedy)
                    clippedx = AAAA.clip((1, 0, 0), invert=False)
                    clippedy = clippedx.clip((0, (- 1), 0), invert=False)
                    c = c.merge(clippedy)
                if (recorte == 0):
                    c = cc



                if (recorte == 2):
                    clippedx = AAAA.clip((1, 0, 0), invert=False)
                    c = c.merge(clippedx)
                p.add_mesh(c, color, opacity=OPA, specular=1, specular_power=15, smooth_shading=True, show_edges=False)
                edges = c.extract_feature_edges(feature_angle=10, boundary_edges=True, feature_edges=False, manifold_edges=False)
                if SYSTEM.SDT[n].Solid_3d_stl != "None" and recorte ==0:
                    print(" ") # No edges
                else:
                    p.add_mesh(edges, 'red')
                points2 = np.c_[(0, 0, 0)]
                c = pv.PolyData(points2)
    p.add_mesh(SYSTEM.DDD, color=[0.5, 0.5, 0.5], opacity=OPA, show_edges=None)
    NN = SYSTEM.AAA.n_blocks
    n = 0
    for g in SYSTEM.side_number:
        if (SYSTEM.SDT[g].Drawing == 1):
            if (SYSTEM.SDT[g].Color == [0, 0, 0]):
                if (SYSTEM.SDT[g].Glass == 'MIRROR'):
                    LL_color = Mirror_color
                else:
                    LL_color = Glass_color
                if (SYSTEM.SDT[g].Glass == 'ABSORB'):
                    LL_color = Absorb_color
                color = LL_color
            else:
                color = SYSTEM.SDT[g].Color
            if (recorte == 1):
                clippedx = SYSTEM.BBB[n].clip('x', invert=False)
                p.add_mesh(clippedx, color ,opacity=OPA, smooth_shading=True, show_edges=None)
                clippedy = SYSTEM.BBB[n].clip('-y', invert=False)
                p.add_mesh(clippedy, color,opacity=OPA, smooth_shading=True, show_edges=None)
            if (recorte == 0):
                No_clipped = SYSTEM.BBB[n]
                p.add_mesh(No_clipped, color,opacity=OPA, smooth_shading=False, show_edges=None)
            if (recorte == 2):
                BBBB = SYSTEM.BBB[n]
                clippedx = BBBB.clip((1, 0, 0), invert=False)
                p.add_mesh(clippedx, color,opacity=OPA, smooth_shading=True, show_edges=None)
        n = (n + 1)








    if INST == True:
       for R in RAYS:
            if (len(R.RayWave) != 0):
                RW = np.asarray(R.RayWave)
                for i in range(0, np.shape(RW)[0]):
                    RGB = wavelength_to_rgb((RW[i] * 1000.0))
                    p.add_mesh(CCC[i], color=RGB, opacity=OPA, smooth_shading=True, line_width=1.0, show_edges=None)

    else:
        if (len(RAYS.RayWave) != 0):
            RW = np.asarray(RAYS.RayWave)
            for i in range(0, np.shape(RW)[0]):
                RGB = wavelength_to_rgb((RW[i] * 1000.0))
                p.add_mesh(CCC[i], color=RGB, opacity=OPA, smooth_shading=True, line_width=1.0, show_edges=None)














             # if (len(RAYS.RayWave) != 0):
             #     RW = np.asarray(RAYS.RayWave)
             #     for i in range(0, np.shape(RW)[0]):
             #         RGB = wavelength_to_rgb((RW[i] * 1000.0))
             #         p.add_mesh(CCC[i], color=RGB, opacity=OPA, smooth_shading=True, line_width=1.0, show_edges=None)



    p.add_axes(line_width=4)

    [cx,cy,cz]=p.center
    p.set_focus([cx,cy,cz])
    # [cpx,cpy,cpz]=p.camera_position
    # print(cpx,cpy,cpz)
    p.camera_position=[-1.0, 0.5, 1.0]

    # print(p.center, " jkjhkjhkh")
    p.set_viewup([0, 1.0, 0])
    p.enable_anti_aliasing()
    # p.disable_anti_aliasing()
    p.enable_eye_dome_lighting()
    p.disable_eye_dome_lighting()
    p.set_background(BackgCol, top=BackgColTop)



    # [wx,wy]=p.window_size
    p.add_text('KrakenOS',position="upper_left" ,font_size=20,color="royalblue")
    p.show_grid(font_size=6,color=GridCol)
    p.show(auto_close=False, interactive=True, interactive_update=False)
    # interactive_update=True esto es para que se pueda continuar trabajando en la ventana

    [cpx,cpy,cpz]=p.camera_position
    # print(cpx,cpy,cpz)


def display2d(SYSTEM, RAYS, view=0, arrow=0):
    """display2d.

    Parameters
    ----------
    SYSTEM :
        SYSTEM
    RAYS :
        RAYS
    view :
        view
    """
    CCC = pv.MultiBlock()
    for rays in RAYS.CC:
        RAY_VTK_OBJ = pv.lines_from_points(rays)
        CCC.append(RAY_VTK_OBJ)




    fs=11
    plt.figure()
    plt.rcParams['font.family'] = 'Arial'
    L_color = [0, 2, 3]
    recorte = view
    NN = SYSTEM.AAA.n_blocks
    if (SYSTEM.SDT[0].Drawing == 0):
        points2 = np.c_[(0.0, 0.0, 0.0)]
        c = pv.PolyData(points2, force_float=False)
    if (SYSTEM.SDT[0].Drawing == 1):
        c = SYSTEM.AAA[0]
    sign=-1.0
    sn=0.65
    mx = []
    mn = []
    for n in range(0, NN):
        sn = sn * sign
        if (SYSTEM.SDT[n].Drawing == 1):
            AAAA = SYSTEM.AAA[n]
            if (SYSTEM.SDT[n].Glass != 'NULL'):
                (PosX, PosY) = SYSTEM.SDT[n].Nm_Pos
                s = SYSTEM.SDT[n].Name
                ss = SYSTEM.Object_Num[n]
                if (view == 0):
                    LT = ''
                    (ax, ay, az) = edge_3d(AAAA, 1, 0, 0)
                    (az, ay) = filter_face_2dplot(az, ay)
                    plt.plot(az, ay, LT, c='black', linewidth=0.5)
                    (ax, ay, az) = edge_3d(AAAA, (- 1), 0, 0)
                    (az, ay) = filter_face_2dplot(az, ay)
                    plt.plot(az, ay, LT, c='black', linewidth=0.5)

                    plt.text(((np.max(az) + PosX) + 1), ((np.max(ay) + PosY) - 1), s, fontsize=fs)
                    delta = ((np.max(ay) - np.min(ay)) / 10)
                    plt.text(az[np.argmin((ay * ay))], sn*1.5*(np.min(ay) - (1.5 * delta)), (('[' + str(ss)) + ']'), fontsize=fs)

                    plt.plot([az[np.argmin((ay * ay))], az[np.argmin((ay * ay))]], [np.mean(ay), sn*1.5*(np.min(ay) - delta)], '-.', c='red', linewidth=0.5)
                    if ((PosX != 0) or (PosY != 0)):
                        plt.arrow((np.max(az) + PosX), (np.max(ay) + PosY/2), (- PosX), (- PosY/2.0), head_width=0.5, head_length=1.0, fc='k', ec='k', length_includes_head=True)
                        plt.arrow((np.max(az) + PosX), (np.max(ay) + PosY), (- PosX)*0, (- PosY)/2.0, head_width=0.1, head_length=0.0, fc='k', ec='k', length_includes_head=True)
                if (view == 1):
                    LT = ''
                    (ax, ay, az) = edge_3d(AAAA, 0, 1, 0)
                    (az, ax) = filter_face_2dplot(az, ax)
                    plt.plot(az, ax, LT, c='black', linewidth=0.5)
                    (ax, ay, az) = edge_3d(AAAA, 0, (- 1), 0)
                    (az, ax) = filter_face_2dplot(az, ax)
                    plt.plot(az, ax, LT, c='black', linewidth=0.5)
                    delta = ((np.max(ax) - np.min(ax)) / 10)

                    plt.text(((np.max(az) + PosX) + 1), ((np.max(ax) + PosY) - 1), s, fontsize=fs)
                    plt.text(az[np.argmin((ax * ax))], sn*1.5*(np.min(ax) - (1.5 * delta)), (('[' + str(ss)) + ']'), fontsize=fs)


                    plt.plot([az[np.argmin((ax * ax))], az[np.argmin((ax * ax))]], [np.mean(ax), sn*1.5*(np.min(ax) - delta)], '-.', c='red', linewidth=0.5)
                    if ((PosX != 0) or (PosY != 0)):
                        plt.arrow((np.max(az) + PosX), (np.max(ax) + PosY/2), (- PosX), (- PosY/2.0), head_width=0.5, head_length=1.0, fc='k', ec='k', length_includes_head=True)
                        plt.arrow((np.max(az) + PosX), (np.max(ax) + PosY), (- PosX)*0, (- PosY)/2.0, head_width=0.1, head_length=0.0, fc='k', ec='k', length_includes_head=True)

            mx.append(np.max(az))
            mn.append(np.min(az))

    NN = SYSTEM.BBB.n_blocks
    for n in range(0, NN):
        TT = SYSTEM.BBB[n]
        sim = '-.'
        if (view == 0):
            (ax, ay, az) = edge_3d(TT, 1, 0, 0)
            plt.plot(az, ay, sim, c='black', linewidth=0.5)
            (ax, ay, az) = edge_3d(AAAA, (- 1), 0, 0)
            plt.plot(az, ay, sim, c='black', linewidth=0.5)
        if (view == 1):
            (ax, ay, az) = edge_3d(TT, 0, 1, 0)
            plt.plot(az, ax, sim, c='black', linewidth=0.5)
            (ax, ay, az) = edge_3d(AAAA, 0, (- 1), 0)
            plt.plot(az, ax, sim, c='black', linewidth=0.5)

    mx = np.asarray(mx)
    mn = np.asarray(mn)

    m1=np.min(mn)
    m2=np.max(mx)
    m3=(m2-m1)/20.0
    plt.plot([m1-m3 , m2+m3],[0,0],'--', color="black", linewidth=1)

    if (len(RAYS.RayWave) != 0):
        RW = np.asarray(RAYS.RayWave)
        for i in range(0, np.shape(RW)[0]):
            RGB = wavelength_to_rgb((RW[i] * 1000.0))
            RRR = CCC[i]
            Ax = RRR.points[:, 0]
            Ay = RRR.points[:, 1]
            Az = RRR.points[:, 2]
            for f in range(1,len(Ax)):
                if (view == 0):
                    line = plt.plot([Az[f-1],Az[f]], [Ay[f-1],Ay[f]], color=RGB, linewidth=0.5)[0]
                if (view == 1):
                    line = plt.plot([Az[f-1],Az[f]], [Ax[f-1],Ax[f]], color=RGB, linewidth=0.5)[0]
                if arrow != 0:
                    add_arrow(line, size=5*arrow)



    plt.title('System Plot')
    plt.xlabel('Z (mm)')
    if (view == 0):
        plt.ylabel('Y (mm)')
    if (view == 1):
        plt.ylabel('X (mm)')
    plt.axis('equal')
    # plt.savefig('plot.pdf')
    plt.show()

def edge_3d(MeshObject, cx, cy, xz):
    """edge_3d.

    Parameters
    ----------
    MeshObject :
        MeshObject
    cx :
        cx
    cy :
        cy
    xz :
        xz
    """
    c = MeshObject.clip((cx, cy, xz), invert=False)
    edges = c.extract_feature_edges(boundary_edges=True, feature_edges=False, manifold_edges=False)
    Ax = edges.points[:, 0]
    Ay = edges.points[:, 1]
    Az = edges.points[:, 2]
    Ax = np.asarray(Ax)
    Ay = np.asarray(Ay)
    Az = np.asarray(Az)
    Xe = []
    Ye = []
    Ze = []
    if (cx == 1):
        i = np.argmin(Ax)
    if (cx == (- 1)):
        i = np.argmax(Ax)
    if (cy == 1):
        i = np.argmin(Ay)
    if (cy == (- 1)):
        i = np.argmax(Ay)
    x0 = Ax[i]
    y0 = Ay[i]
    z0 = Az[i]
    for j in range(0, (np.shape(Ax)[0] - 1)):
        AAx = Ax[i]
        AAy = Ay[i]
        AAz = Az[i]
        Xe.append(AAx)
        Ye.append(AAy)
        Ze.append(AAz)
        Ax = np.delete(Ax, i)
        Ay = np.delete(Ay, i)
        Az = np.delete(Az, i)
        X = (Ax - AAx)
        Y = (Ay - AAy)
        Z = (Az - AAz)
        R = np.sqrt((((X * X) + (Y * Y)) + (Z * Z)))
        i = np.argmin(R)
    Xe.append(x0)
    Ye.append(y0)
    Ze.append(z0)
    Xe = np.asarray(Xe)
    Ye = np.asarray(Ye)
    Ze = np.asarray(Ze)
    return (Xe, Ye, Ze)

def filter_face_2dplot(v1, v2):
    """filter_face_2dplot.

    Parameters
    ----------
    v1 :
        v1
    v2 :
        v2
    """
    av1 = np.copy(v1)
    av2 = np.copy(v2)
    av1 = np.roll(av1, (- 1))
    av2 = np.roll(av2, (- 1))
    yy = (v1 - av1)
    zz = (v2 - av2)
    R = np.sqrt(((yy * yy) + (zz * zz)))
    M = np.mean(R)
    AW = np.argwhere((R < M))
    v1 = v1[AW]
    v2 = v2[AW]
    av1 = np.copy(v1)
    av2 = np.copy(v2)
    av1 = np.roll(av1, 1)
    av2 = np.roll(av2, 1)
    yy = (v1 - av1)
    zz = (v2 - av2)
    R = np.sqrt(((yy * yy) + (zz * zz)))
    AW = np.argmax(R)
    v1 = np.roll(v1, (- AW))
    v2 = np.roll(v2, (- AW))
    return (v1, v2)




# """ change pip to True for pip installation """

# pip= False


# if pip == False:
#     import csv
#     import numpy as np

#     def load_alluminum_complex(file):
#         """load_alluminum_complex.

#         Parameters
#         ----------
#         file :
#             file
#         """
#         w = []
#         n = []
#         k = []
#         with open(file, 'rt') as f:
#             csv_reader = csv.reader(f, delimiter=';')
#             next(csv_reader)
#             for line in csv_reader:
#                 w.append(float(line[0]))
#                 n.append(float(line[1]))
#                 k.append(float(line[2]))
#         return (w, n, k)

#     def load_Catalog(FileCat):
#         """load_Catalog.

#         Parameters
#         ----------
#         FileCat :
#             FileCat
#         """
#         cat = []
#         ARR_CAT = []
#         ARR_NM = []
#         ARR_ED = []
#         ARR_CD = []
#         ARR_TD = []
#         ARR_OD = []
#         ARR_LD = []
#         ARR_IT = []
#         print('Loading glass calatogs:')
#         for file in FileCat:
#             ARR_CAT.append(file)
#             f = open(file, 'r')
#             for x in f:
#                 cat.append(x)
#         con = 0
#         coords = []
#         names = []
#         for f in cat:
#             cadena = f.split()
#             cad = cadena[0]
#             if (cad == 'NM'):
#                 coords.append(con)
#                 names.append(cadena[1])
#             con = (con + 1)
#         names = np.asarray(names)
#         for i in range(0, (len(coords) - 1)):
#             ITT = []
#             for j in range(coords[i], coords[(i + 1)]):
#                 cadena = cat[j].split()
#                 cad = cat[j][2:].split()
#                 if (cadena[0] == 'NM'):
#                     NM = cad
#                 if (cadena[0] == 'ED'):
#                     ED = cad
#                 if (cadena[0] == 'CD'):
#                     CD = cad
#                 if (cadena[0] == 'TD'):
#                     TD = cad
#                 if (cadena[0] == 'OD'):
#                     OD = cad
#                 if (cadena[0] == 'LD'):
#                     LD = cad
#                 if (cadena[0] == 'IT'):
#                     IT = cad
#                     ITT.append(IT)
#             NM = np.asarray(NM[1:(- 1)], dtype=np.float64)
#             ED = np.asarray(ED, dtype=np.float64)
#             CD = np.asarray(CD, dtype=np.float64)
#             TD = np.asarray(TD, dtype=np.float64)
#             OD = np.asarray(OD)
#             LD = np.asarray(LD, dtype=np.float64)
#             IT = np.asarray(ITT, dtype=np.float64).T
#             ARR_NM.append(NM)
#             ARR_ED.append(ED)
#             ARR_CD.append(CD)
#             ARR_TD.append(TD)
#             ARR_OD.append(OD)
#             ARR_LD.append(LD)
#             ARR_IT.append(IT)
#         CATALOG = [ARR_CAT, names, ARR_NM, ARR_ED, ARR_CD, ARR_TD, ARR_OD, ARR_LD, ARR_IT]
#         return CATALOG



# else:

import csv
import numpy as np

def load_alluminum_complex(file):
    """load_alluminum_complex.

    Parameters
    ----------
    file :
        file
    """
    w = []
    n = []
    k = []
    with open(file, 'rt') as f:
        csv_reader = csv.reader(f, delimiter=';')
        next(csv_reader)
        for line in csv_reader:
            w.append(float(line[0]))
            n.append(float(line[1]))
            k.append(float(line[2]))
    return (w, n, k)

def load_Catalog(FileCat):
    """load_Catalog.

    Parameters
    ----------
    FileCat :
        FileCat
    """
    cat = []
    ARR_CAT = []
    ARR_NM = []
    ARR_ED = []
    ARR_CD = []
    ARR_TD = []
    ARR_OD = []
    ARR_LD = []
    ARR_IT = []
    print('Processing glass catalog:')
    for file in FileCat:
        ARR_CAT.append(file)
        f = open(file, 'r')
        for x in f:

            if not x.isspace():
                 cat.append(x)

            # cat.append(x)
    con = 0
    coords = []
    names = []
    for f in cat:
        cadena = f.split()
        cad = cadena[0]
        if (cad == 'NM'):
            coords.append(con)
            names.append(cadena[1])
        con = (con + 1)
    names = np.asarray(names)
    for i in range(0, (len(coords) - 1)):
        ITT = []
        for j in range(coords[i], coords[(i + 1)]):
            cadena = cat[j].split()
            cad = cat[j][2:].split()
            if (cadena[0] == 'NM'):
                NM = cad
            if (cadena[0] == 'ED'):
                ED = cad
                if ED[1]=="-":
                    ED[1]="0.0"
            if (cadena[0] == 'CD'):
                CD = cad
            if (cadena[0] == 'TD'):
                TD = cad
            if (cadena[0] == 'OD'):
                OD = cad
            if (cadena[0] == 'LD'):
                LD = cad
            if (cadena[0] == 'IT'):
                IT = cad
                if len(IT)==3:
                    ITT.append(IT)

        NM = np.asarray(NM[1:(- 1)], dtype=np.float64)

        # print(names[i])
        # print(ED)

        ED = np.asarray(ED, dtype=np.float64)
        CD = np.asarray(CD, dtype=np.float64)
        TD = np.asarray(TD, dtype=np.float64)
        OD = np.asarray(OD)
        LD = np.asarray(LD, dtype=np.float64)
        IT = np.asarray(ITT, dtype=np.float64).T
        ARR_NM.append(NM)
        ARR_ED.append(ED)
        ARR_CD.append(CD)
        ARR_TD.append(TD)
        ARR_OD.append(OD)
        ARR_LD.append(LD)
        ARR_IT.append(IT)
    CATALOG = [ARR_CAT, names, ARR_NM, ARR_ED, ARR_CD, ARR_TD, ARR_OD, ARR_LD, ARR_IT]
    return CATALOG



import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import scipy
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

P_Obj = Kos.surf()

P_Obj.Thickness = 5000.0
P_Obj.Glass = "AIR"
P_Obj.Diameter = 6.796727741707513E+002 * 2.0
P_Obj.Drawing = 0

######################################################################################


objeto = Kos.surf()
objeto.Rc = -12000
objeto.k = -1
objeto.Diameter = 2500
objeto.Thickness = -6000
objeto.Glass = "MIRROR"

P_Ima = Kos.surf()
P_Ima.Rc = 0
P_Ima.Thickness = -1.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 6000.0
P_Ima.Drawing = 1
P_Ima.Name = "Plano imagen"

A = [P_Obj, objeto, P_Ima]

configur = Kos.Setup()
Telescope = Kos.system(A, configur)
Rays = Kos.raykeeper(Telescope)

W = 0.633
Sun = Kos.SourceRnd()

examp = 4
if examp == 0:
    # Sun distribution
    def f(x):
        teta = x
        FI = np.zeros_like(teta)
        Arg2 = np.argwhere(teta > (4.65 / 1000.0))
        FI = np.cos(0.326 * teta) / np.cos(0.308 * teta)

        Chi2 = .03
        k = 0.9 * np.log(13.5 * Chi2) * np.power(Chi2, -0.3)
        r = (2.2 * np.log(0.52 * Chi2) * np.power(Chi2, 0.43)) - 1.0
        FI[Arg2] = np.exp(k) * np.power(teta[Arg2] * 1.0e3, r)
        return FI


    Sun.field = 20 * np.rad2deg((4.65 / 1000.0))

if examp == 1:
    # Sinc cunction
    def f(x):
        Wh = 0.025
        r = (x * 90.0 / 0.025) * np.pi
        res = np.sin(r) / r
        return res


    Sun.field = 0.025 * 3

if examp == 2:
    # Flat
    def f(x):
        res = 1
        return res


    Sun.field = 1.2 / (2. * 3600.)

if examp == 3:
    # Parabolic
    def f(x):
        r = (x * 90.0 / 0.025)
        res = r ** 2
        return res


    Sun.field = 1.2 / (2. * 3600.)

if examp == 4:
    # Gaussian (Seeing)
    def f(x):
        x = np.rad2deg(x)
        seing = 1.2 / 3600.0
        sigma = seing / 2.3548
        mean = 0
        standard_deviation = sigma
        y = scipy.stats.norm(mean, standard_deviation)
        res = y.pdf(x)
        return res


    Sun.field = 4 * 1.2 / (2.0 * 3600.0)

Sun.fun = f
Sun.dim = 3000
Sun.num = 100000
L, M, N, X, Y, Z = Sun.rays()

Xr = np.zeros_like(L)
Yr = np.zeros_like(L)
Nr = np.zeros_like(L)

con = 0
con2 = 0
for i in range(0, Sun.num):
    if con2 == 10:
        print(100. * i / Sun.num)
        con2 = 0

    pSource_0 = [X[i], Y[i], Z[i]]
    dCos = [L[i], M[i], N[i]]
    Telescope.Trace(pSource_0, dCos, W)

    Xr[con] = Telescope.Hit_x[-1]
    Yr[con] = Telescope.Hit_y[-1]
    Nr[con] = Telescope.SURFACE[-1]
    con = con + 1
    con2 = con2 + 1
    # Rays.push()

args = np.argwhere(Nr == 2)
plt.plot(Xr[args], Yr[args], '.', c="g", markersize=1)

# axis labeling
plt.xlabel('x')
plt.ylabel('y')

# figure name
plt.title('Spot diagram')
plt.axis('square')
plt.show()

#             Rays.push()
# Kos.display3d(Telescope, Rays, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Extra Shape Micro Lens Array"""

import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
import numpy as np
import matplotlib.pyplot as plt

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 10
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0

# ______________________________________#

L1a = Kos.surf()
L1a.Rc = 55.134 * 0
L1a.Thickness = 2.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0

# ______________________________________#

L1c = Kos.surf()
L1c.Thickness = 40
L1c.Glass = "AIR"
L1c.Diameter = 30


# ______________________________________#

def f(x, y, E):
    DeltaX = E[0] * np.rint(x / E[0])
    DeltaY = E[0] * np.rint(y / E[0])
    x = x - DeltaX
    y = y - DeltaY
    s = np.sqrt((x * x) + (y * y))
    c = 1.0 / E[1]
    InRoot = 1 - (E[2] + 1.0) * c * c * s * s
    z = (c * s * s / (1.0 + np.sqrt(InRoot)))
    return z


# ______________________________________#

coef = [3.0, -3, 0]
L1c.ExtraData = [f, coef]
L1c.Res = 2

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 300.0
P_Ima.Name = "Image plane"

# ______________________________________#

A = [P_Obj, L1a, L1c, P_Ima]
Config_1 = Kos.Setup()

# ______________________________________#

Lens = Kos.system(A, Config_1)
Rays = Kos.raykeeper(Lens)

# ______________________________________#

Wav = 0.45
for i in range(-100, 100 + 1):
    pSource = [0.0, i / 10., 0.0]
    dCos = [0.0, 0.0, 1.0]
    Lens.Trace(pSource, dCos, Wav)
    Rays.push()

# ______________________________________#

Kos.display3d(Lens, Rays, 1)
# Kos.display2d(Lens, Rays, 0)



#!/usr/bin/env Python3
import os

from .ciddorModel import *
from .dispersionModels import *
from .refractionModels import *
from .refractionModels import *
from .misc import *
from .refractivityModels import *
from .neoslalib import *
from .observatories import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Spyder Spot Diagram"""

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1.059E+003 * 2.0
P_Obj.Drawing = 0

# ______________________________________#

Spider = Kos.surf()
Spider.Rc = 999999999999.0
Spider.Thickness = 3.452229924716749E+003 + 100.0
Spider.Glass = "AIR"
Spider.Diameter = 1.059E+003 * 2.0
plane1 = pv.Plane(center=[0, 0, 0], direction=[0, 0, 1], i_size=30, j_size=2100, i_resolution=10, j_resolution=10)
plane2 = pv.Plane(center=[0, 0, 0], direction=[0, 0, 1], i_size=2100, j_size=30, i_resolution=10, j_resolution=10)
Baffle1 = pv.Disc(center=[0.0, 0.0, 0.0], inner=0, outer=875 / 2.0, normal=[0, 0, 1], r_res=1, c_res=100)
Baffle2 = Baffle1.merge(plane1)
Baffle3 = Baffle2.merge(plane2)
AAA = pv.MultiBlock()
AAA.append(plane1)
AAA.append(plane2)
AAA.append(Baffle1)
Spider.Mask_Shape = AAA
Spider.Mask_Type = 2
Spider.TiltZ = 0

# ______________________________________#

Thickness = 3.452229924716749E+003
M1 = Kos.surf()
M1.Rc = -9.638000000004009E+003
M1.Thickness = -Thickness
M1.k = -1.077310000000000E+000
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0

# ______________________________________#

M2 = Kos.surf()
M2.Rc = -3.93E+003
focusShift = 1.0  # Set cero to focus, 1 is only to see the spider is the spot diagram
M2.Thickness = Thickness + 1.037179115116706E+003 + focusShift
M2.k = -4.328100000000000E+000
M2.Glass = "MIRROR"
M2.Diameter = 3.365E+002 * 2.0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 100.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, Spider, M1, M2, P_Ima]
configuracion_1 = Kos.Setup()

# ______________________________________#

Telescopio = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Telescopio)

# ______________________________________#

tam = 7
rad = 2200 / 2
tsis = len(A) - 1
for i in range(-tam, tam + 1):
    for j in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            W = 0.6
            Telescopio.Trace(pSource_0, dCos, W)
            Rayos.push()

# ______________________________________#

Kos.display3d(Telescopio, Rayos, 2)
X, Y, Z, L, M, N = Rayos.pick(-1)

# ______________________________________#

plt.plot(X, Y, 'x')
plt.xlabel('x')
plt.ylabel('y')
plt.title('Spot Diagram')
plt.axis('square')
plt.show()



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Wavefront Fitting"""

import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an library downloaded from github is running"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

import scipy



###########-----------------------




import matplotlib.pyplot as plt
def MyPlot(RK,surf, figure= "Spot", mk=["x"], col = [[0.8,0.0,0.0]]):
    fig = plt.figure(figure)
    ax = fig.add_subplot()
    i=0
    for rk in RK:
        x,y,z,l,m,n = rk.pick(surf[i], coordinates="local")
        plt.scatter(x, y,marker = mk[i], color = col[i])
        i = i + 1
    # https://matplotlib.org/3.1.1/api/markers_api.html
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel('x (mm)')
    ax.set_ylabel('y (mm)')
    ax.set_title('Spot diagram')
    plt.show()



def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS


def BestRMS_2(system,raykeeper):
    x,y,z,l,m,n = raykeeper.pick(-1 , coordinates="local")
    rms = R_RMS_delta(0,l,m,n,x,y)
    return rms


def BestFocus(X, Y, Z, L, M, N, system, mod=1):
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    if mod ==1:
        system.SDT[-2].Thickness = system.SDT[-2].Thickness + v[0]
        system.SetData()
        system.SetSolid()
    return system, v[0]

class Function2Optimize:
    def __init__(self, system, raykeeper, effl, Surf, w, AperVal, AperType):

        self.system = system
        self.w = w
        self.effl = effl
        self.raykeeper = raykeeper

        self.P = Kos.PupilCalc(system, Surf, w, AperType, AperVal)

        self.P.Samp, self.P.Ptype, self.P.FieldType = 25, "fanx", "angle"
        self.P.FieldX, self.P.FieldY = 0.0, 0.0
        self.ProcessPattern2Field()

    def ProcessPattern2Field(self):
        self.x, self.y, self.z, self.L, self.M, self.N = self.P.Pattern2Field()

    def EFFL(self, V):
        self.system.SDT[1].k = V[0]
        self.system.SDT[2].k = V[1]

        self.system.Parax(self.w)
        D_EFFL = self.system.EFFL - self.effl

        Kos.TraceLoop(self.x, self.y, self.z, self.L, self.M, self.N, self.w, \
                      self.raykeeper, clean = 1)

        B = BestRMS_2(self.system, self.raykeeper)
        self.system.RestoreData()
        return [D_EFFL, B]





class Function2OptimizeSeidel:
    def __init__(self, ABER, w, effl):
        self.ABER = ABER
        self.w =w
        self.effl = effl

    def Fun(self, V):
        System = self.ABER.SYSTEM
        System.SDT[1].k = V[0]
        System.SDT[2].k = V[1]


        System.SetData()

        self.ABER.calculate()

        System.RestoreData()
        Sph = self.ABER.SAC_TOTAL[0]
        Coma = self.ABER.SAC_TOTAL[1]
        Ast = self.ABER.SAC_TOTAL[2]

        return [Sph, Coma]
# ______________________________________#

currentDirectory = os.getcwd()
sys.path.insert(1, currentDirectory + '/library')

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Thickness = 2000.0
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1300.0
P_Obj.Drawing = 0

Thickness = 1750.6548739435177
M1 = Kos.surf()
M1.Rc = -6034.3700365797185
M1.Thickness = -Thickness
M1.k = -1.0
M1.Glass = "MIRROR"
M1.Diameter = 1300
M1.InDiameter = 650.0 * 0.5

M2 = Kos.surf()
M2.Rc = -6034.3700365797185
M2.Thickness = Thickness
M2.k = 0.0
M2.Glass = "MIRROR"
M2.Diameter = 650.0

M1Vertex=Kos.surf(Thickness = 20.0, Glass = "AIR",Diameter = 600.0)
M1Vertex.Drawing=0

Corr_1=Kos.surf(Thickness = 10, Glass = "LITHOSIL-Q", Diameter=230)

ED1=np.zeros(20)
ED1[0] = 0.0
ED1[1] = 0.0
Corr_1.AspherData = ED1

Corr_2=Kos.surf(Thickness = 420, Glass="AIR", Diameter=230)
Corr_2.AspherData = -ED1

P_Ima = Kos.surf()
P_Ima.Diameter = 700.0

A = [P_Obj, M1, M2, M1Vertex, Corr_1, Corr_2, P_Ima]
configuracion_1 = Kos.Setup()
Telescope = Kos.system(A, configuracion_1)
# ______________________________________#


Rays = Kos.raykeeper(Telescope)
Rays0 = Kos.raykeeper(Telescope)
Rays1 = Kos.raykeeper(Telescope)
Rays2 = Kos.raykeeper(Telescope)


W = 0.6
Surf=1
AperVal = 1300.
AperType = "EPD"
Pupil = Kos.PupilCalc(Telescope, Surf, W, AperType, AperVal)
Pupil.Samp = 10
Pupil.Ptype = "hexapolar"
Pupil.FieldX = 1.7/2.0
Pupil.FieldY = 0.0

AB = Kos.Seidel(Pupil)
effl = 5200

MyFun = Function2OptimizeSeidel(AB, W, effl)

K1 = -1.
K2 = 0



LimInf = [K1 - 3, K2 - 60]
LimSup = [K1 + 3, K2 + 60]
b=(LimInf, LimSup)

R = scipy.optimize.least_squares(MyFun.Fun, [K1, K2],bounds=b, verbose=2)

[K1, K2] = R.x

Telescope.SDT[1].k = K1
Telescope.SDT[2].k = K2



Telescope.SetData()
Telescope.SetSolid()


Surf, W, AperVal, AperType = 1, W, M1.Diameter, "STOP"
P = Kos.PupilCalc(Telescope, Surf, W, AperType, AperVal)
P.Samp, P.Ptype, P.FieldX, P.FieldType = 11, "hexapolar", (1.7 / 2.0) * 0.707, "angle"
x, y, z, L, M, N = P.Pattern2Field()

DeltaW = 0.1
Kos.TraceLoop(x, y, z, L, M, N, W, Rays, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W + DeltaW, Rays, clean = 0)
Kos.TraceLoop(x, y, z, L, M, N, W - DeltaW, Rays, clean = 0)
x1, y1, z1, l1, m1, n1 = Rays.pick(-1 , coordinates="local")

BestFocus(x1,y1,z1,l1,m1,n1, Telescope, mod=1)





""" -----------------------------------------"""



P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "hexapolar", 0, "angle"
x, y, z, L, M, N = P.Pattern2Field()


Kos.TraceLoop(x, y, z, L, M, N, W, Rays0, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W + DeltaW, Rays1, clean = 0)
Kos.TraceLoop(x, y, z, L, M, N, W - DeltaW, Rays2, clean = 0)
x, y, z, l, m, n = Rays0.pick(-1 , coordinates="local")


rms = R_RMS_delta(0,l,m,n,x,y)
print("K1: ", K1, " K2: ", K2, " RMS: ", rms )

# Kos.display3d(Telescope, Rays, 2)


RAYS = [Rays0, Rays1, Rays2]
MK = ["x", "*", "+"]
COL = [[0.8,0.0,0.0], [0.0,0.8,0.0], [0.0,0.0,0.8]]
SURF = [-1, -1, -1]
MyPlot(RAYS, SURF, figure= "Spot1", mk = MK, col = COL)


""" ---------------------------------------------"""

P.Samp, P.Ptype, P.FieldX, P.FieldType = 7, "hexapolar", 1.7/2, "angle"
x, y, z, L, M, N = P.Pattern2Field()

Kos.TraceLoop(x, y, z, L, M, N, W, Rays0, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W + DeltaW, Rays1, clean = 1)
Kos.TraceLoop(x, y, z, L, M, N, W - DeltaW, Rays2, clean = 1)
x, y, z, l, m, n = Rays0.pick(-1 , coordinates="local")
system, deltaZ = BestFocus(x, y, z, l, m, n, Telescope, mod = 1)

rms = R_RMS_delta(0,l,m,n,x,y)
print("K1: ", K1, " K2: ", K2, " RMS: ", rms )


RAYS = [Rays0, Rays1, Rays2]
MK = ["x", "*", "+"]
COL = [[0.8,0.0,0.0], [0.0,0.8,0.0], [0.0,0.0,0.8]]
SURF = [-1, -1, -1]
MyPlot(RAYS, SURF, figure= "Spot2", mk = MK, col = COL)


Telescope.Parax(W)
print(Telescope.EFFL)

Kos.display3d(Telescope, Rays, 1)
Telescope.Parax(W)
print(Telescope.EFFL)













W = 0.6
Surf=1
AperVal = 1300.
AperType = "STOP"
Pupil = Kos.PupilCalc(Telescope, Surf, W, AperType, AperVal)
Pupil.Samp = 10
Pupil.Ptype = "hexapolar"
Pupil.FieldX = 1.7/2.0
Pupil.FieldY = 0.0

AB = Kos.Seidel(Pupil)
print("--------------------------------------")
print(AB.SCW_AN)
print(AB.SCW_NM)
print(AB.SCW_TOTAL)




import numpy as np
import matplotlib.pyplot as plt
import KrakenOS as Kos

def ZernikeDataImage2Plot(datos, Type='interferogram'):
    """ZernikeDataImage2Plot.

    Parameters
    ----------
    datos :
        datos
    Type :
        Type
    """
    if (Type == 'interferogram'):


        datos = np.sin(((2 * np.pi) * datos))

    if (Type == 'phase'):
        datos = datos
    img = (datos / np.max(datos))
    img = (img * 255)
    img = img.astype(int)
    img = np.asarray(img)
    plt.figure(987)
    plt.imshow(img, cmap='gray')
    plt.show()
    return 0

def WavefrontData2Image(z_coeff, res=323):
    """WavefrontData2Image.

    Parameters
    ----------
    z_coeff :
        z_coeff
    res :
        res
    """
    TamImag = int(res)
    r = (TamImag / 2.0)
    ARRAY_ZERNIKE = np.zeros((TamImag, TamImag))
    H = []
    K = []
    X = []
    Y = []
    for h in range(0, TamImag):
        for k in range(0, TamImag):
            x = ((h - r) / r)
            y = ((k - r) / r)
            RP = np.sqrt(((x * x) + (y * y)))
            if (RP <= 1):
                H.append(h)
                K.append(k)
                X.append(x)
                Y.append(y)
    H = np.asarray(H)
    K = np.asarray(K)
    X = np.asarray(X)
    Y = np.asarray(Y)
    Z = Kos.Wavefront_Zernike_Phase(X, Y, z_coeff)
    ARRAY_ZERNIKE[(H, K)] = Z
    return ARRAY_ZERNIKE




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Doublet Lens Commands System"""

import numpy as np
import pkg_resources
""" Looking for if KrakenOS is installed, if not, it assumes that
an folder downloaded from github is run"""

required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("Not installed")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# _________________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 0.1
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.

# _________________________________________#

P_Obj2 = Kos.surf()
P_Obj2.Rc = 0.0
P_Obj2.Thickness = 10
P_Obj2.Glass = "AIR"
P_Obj2.Diameter = 30.0

# _________________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0
L1a.Axicon = 0

# _________________________________________#

L1b = Kos.surf()
L1b.Rc = -3.071608670000159E+001
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# _________________________________________#

L1c = Kos.surf()
L1c.Rc = -7.819730726078505E+001
L1c.Thickness = 9.737604742910693E+001
L1c.Glass = "AIR"
L1c.Diameter = 30

# _________________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 18.0
P_Ima.Name = "Plano imagen"

# _________________________________________#

A = [P_Obj, P_Obj2, L1a, L1b, L1c, P_Ima]
configuracion_1 = Kos.Setup()

# _________________________________________#

Doblete = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Doblete)

# _________________________________________#

pSource_0 = [0, 14, 0]
tet = 0.1
dCos = [0.0, np.sin(np.deg2rad(tet)), -np.cos(np.deg2rad(tet))]
W = 0.4
Doblete.Trace(pSource_0, dCos, W)
Rayos.push()

# _________________________________________#

Kos.display3d(Doblete, Rayos, 2)

# _________________________________________#

print("Distancia focal efectiva")
print(Doblete.EFFL)
print("Plano principal anterior")
print(Doblete.PPA)
print("Plano principal posterior")
print(Doblete.PPP)
print("Superficies tocadas por el rayo")
print(Doblete.SURFACE)
print("Nombre de la superficie")
print(Doblete.NAME)
print("Vidrio de la superficie")
print(Doblete.GLASS)
print("Coordenadas del rayo en las superficies")
print(Doblete.XYZ)
print("Etc, ver documentaciòn")
print(Doblete.S_XYZ)
print(Doblete.T_XYZ)
print(Doblete.OST_XYZ)
print(Doblete.DISTANCE)
print(Doblete.OP)
print(Doblete.TOP)
print(Doblete.TOP_S)
print(Doblete.ALPHA)
print(Doblete.S_LMN)
print(Doblete.LMN)
print(Doblete.R_LMN)
print(Doblete.N0)
print(Doblete.N1)
print(Doblete.WAV)
print(Doblete.G_LMN)
print(Doblete.ORDER)
print(Doblete.GRATING)
print(Doblete.RP)
print(Doblete.RS)
print(Doblete.TP)
print(Doblete.TS)
print(Doblete.TTBE)
print(Doblete.TT)
print(Doblete.BULK_TRANS)



# from numba import jit

import numpy as np
import os
import sys
import random
import inspect

from .SurfTools import surface_tools as SUT
from .Prerequisites3D import *

from .Physics import *
from .HitOnSurf import *
from .InterNormalCalc import *

import timeit



# from .__init__ import RUTE as ModRute

def prob(pro):
    """prob.

    Parameters
    ----------
    pro :
        pro
    """
    a_list = [0, 1]
    prob = pro
    distribution = [prob, (1.0 - prob)]
    random_number = random.choices(a_list, distribution)
    return random_number

class system():


    # print(ModRute)

    """system.
            SYSTEM CLASS ATRIBUTES AND IMPLEMENTATIONS:


          system.Trace (pS, dC, wV)
             Sequential ray tracing.
             pS = [1.0, 0.0, 0.0] – Ray origin coordinates
             dC = [0.0,0.0,1.0] - The directing cosines
             wV = 0.4 - Wavelength


          system.NsTrace (pS, dC, wV)
             Non-Sequential ray tracing


          system.Parax (w)
             Paraxial optics calculations


          system.disable_inner
             Enables the central aperture.


          system.enable_inner
             Disables the central aperture.


          system.SURFACE
	    Returns the surfaces the ray passed through.


          system.NAME
        	    Returns surface names that the ray passed through


          system.GLASS
        	    Returns materials that the ray passed through.


          system.XYZ
             [X, Y, Z] ray coordinates from its origin to the image plane.


          system.OST_XYZ
             [X, Y, Z] coordinates of ray intersections in reference to
            a coordinate system at its vertex, even if this vertex has
            a translation or rotation.


          system.DISTANCE
             List of distances traveled by the ray.


          system.OP
        	List of optical paths.


          system.TOP
        	Total optical path.


          system.TOP_S     List of the ray's optical path by sections.


          system.ALPHA
             List the materials absorption coefficients


          system.BULK_TRANS
             List the transmission through all the system absorption
             coefficients are considered.


          system.S_LMN
             Surface normal direction cosines [L, M, N].


          system.LMN
             incident ray direction cosines [L, M, N].


          system.R_LMN
             Resulting ray direction cosines [L, M, N].


          system.N0
             Refractive indices before and after each interface


          system.N1
             Refractive indices after each interface.
             This is useful to differentiate between index
             before and after an iteration. Example:

             N0 = [n1, n2, n3, n4, n5]
             N1 = [n2, n3, n4, n5, n5]


          system.WAV
        	    Wavelength of the ray (µm)


          system.G_LMN
             [L, M, N] Direction cosines that define the lines
             on the diffraction grating on the plane.


          system.ORDER
        	     Ray diffraction order.


          system.GRATING_D
	     Distance between lines of the diffraction grating.
             Units (Microns)


          system.RP
	     Fresnel reflection coefficients for polarization P.


          system.RS
             Fresnel reflection coefficients for polarization S.


          system.TP
            Fresnel transmission coefficients for polarization P.


          system.TS
            Fresnel transmission coefficients for polarization S.


          system.TTBE
            Total energy transmitted or reflected per element.


          system.TT
            Total energy transmitted or reflected total.


          system.targ_surf (int)
            Limits the ray tracing to the defined surface


          system.flat_surf (int)
            Change a surface to flat.

    """


    def __init__(self, SurfData, KN_Setup):
        """__init__.

        Parameters
        ----------
        SurfData :
            SurfData
        KN_Setup :
            KN_Setup
        """
        self.ExectTime=[]

        self.SDT = SurfData
        self.update = False
        self.S_Matrix = []
        self.N_Matrix = []
        self.SistemMatrix = np.matrix([[1.0, 0.0], [0.0, 1.0]])
        (self.a, self.b, self.c, self.d, self.EFFL, self.PPA, self.PPP) = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
        self.SETUP = KN_Setup
        self.n = len(self.SDT)
        for ii in range(0, self.n):
            self.SDT[ii].SaveSetup()
        self.SuTo = SUT(self.SDT)
        self.Object_Num = np.arange(0, self.n, 1)
        self.Targ_Surf = self.n
        self.SuTo.Surface_Flattener = 0
        self.Disable_Inner = 1
        self.ExtraDiameter = 0
        self.SuTo.ErrSurfCase = 0
        self.__SurFuncSuscrip()
        self.Pr3D = Prerequisites(self.SDT, self.SuTo)
        self.__PrerequisitesGlass()
        self.Pr3D.Prerequisites3SMath()
        self.Pr3D.Prerequisites3DSolids()
        self.PreWave = (- 1e-09)
        self.AAA = self.Pr3D.AAA
        self.BBB = self.Pr3D.BBB
        self.DDD = self.Pr3D.DDD
        self.EEE = self.Pr3D.EEE
        self.GlassOnSide = self.Pr3D.GlassOnSide
        self.side_number = self.Pr3D.side_number
        self.TypeTotal = self.Pr3D.TypeTotal
        self.TRANS_1A = self.Pr3D.TRANS_1A
        self.TRANS_2A = self.Pr3D.TRANS_2A
        self.HS = Hit_Solver(self.SDT)
        self.INORM = InterNormalCalc(self.SDT, self.TypeTotal, self.Pr3D, self.HS)
        self.INORM.Disable_Inner = 1
        self.Pr3D.Disable_Inner = 1
        (self.c_p, self.n_p, self.d_p) = (0, 0, 0)
        self.tt=1.
        self.energy_probability = 0
        self.NsLimit = 200
        self.ang = 0


        # (ResVec, CurrN, sign ,ang) = self.SDT[2].PHYSICS.calculate([-0.01295049, 0.14862896 ,0.98880823], [-0., -0., -1.], 1.0, 1.0, [0., 1. ,0.], 0.0, 0.0, 0.6, 0.)

    def __SurFuncSuscrip(self):
        """__SurFuncSuscrip.
        """
        for i in range(0, self.n):
            self.SDT[i].build_surface_function()

    def __PrerequisitesGlass(self):
        """__PrerequisitesGlass.
        """
        self.Glass = []
        self.GlobGlass = []
        for i in range(0, self.n):
            if (type(self.SDT[i].Glass) == type(1.0)):
                self.SDT[i].Glass = ('AIR_' + str(self.SDT[i].Glass))
                # print(self.SDT[i].Glass)
            self.Glass.append(self.SDT[i].Glass.replace(' ', ''))
            self.GlobGlass.append(self.SDT[i].Glass.replace(' ', ''))
            if ((self.GlobGlass[i] == 'NULL') or (self.GlobGlass[i] == 'ABSORB')):
                self.GlobGlass[i] = self.GlobGlass[(i - 1)]

    def __NonSequentialChooserToot(self, A_RayOrig, A_Proto_pTarget, k):
        """__NonSequentialChooserToot.

        Parameters
        ----------
        A_RayOrig :
            A_RayOrig
        A_Proto_pTarget :
            A_Proto_pTarget
        k :
            k
        """
        ng = self.GlassOnSide[k]
        A_Glass = self.SDT[ng].Glass
        if (A_Glass == 'NULL'):
            distance = 99999999999999.9
        else:
            (A_pTarget, A_SurfHit) = self.EEE[k].ray_trace(A_RayOrig, A_Proto_pTarget)
            if (len(A_SurfHit) == 0):
                distance = 99999999999999.9
            else:
                s = 0
                h = []
                for f in A_SurfHit:
                    PD = (np.asarray(A_pTarget[s]) - np.asarray(A_RayOrig))
                    distance = np.linalg.norm(PD)
                    if (np.abs(distance) < 0.05):
                        distance = 99999999999999.9
                    h.append(distance)
                    s = (s + 1)
                distance = np.min(np.asarray(h))
        return distance

    def __NonSequentialChooser(self, SIGN, A_RayOrig, ResVec, j):
        """__NonSequentialChooser.

        Parameters
        ----------
        SIGN :
            SIGN
        A_RayOrig :
            A_RayOrig
        ResVec :
            ResVec
        j :
            j
        """
        chooser = []
        [SLL, SMM, SNN] = ResVec
        distance = 99999999999999.9
        [Px, Py, Pz] = A_RayOrig
        [LL, MM, NN] = ResVec
        A_Proto_pTarget = (np.asarray(A_RayOrig) + ((np.asarray(ResVec) * 999999999.9) * SIGN))
        for k in range(1, len(self.EEE)):
            distance = self.__NonSequentialChooserToot(A_RayOrig, A_Proto_pTarget, k)
            chooser.append(distance)
        chooser = np.asarray(chooser)
        jj = (np.argmin(chooser) + 1)
        chooser[(jj - 1)] = 99999999999999.9
        kk = (np.argmin(chooser) + 1)
        (A_pTarget, A_SurfHit) = self.EEE[int(jj)].ray_trace(A_RayOrig, A_Proto_pTarget)
        PRR = np.shape(A_SurfHit)[0]
        return (int(j), int(jj), int(kk), PRR)

    def __CollectDataInit(self):
        """__CollectDataInit.
        """
        self.val = 1
        self.SURFACE = []
        self.NAME = []
        self.GLASS = []
        self.S_XYZ = []
        self.T_XYZ = []
        self.XYZ = []
        self.XYZ.append([0, 0, 0])

        self.OST_XYZ = []
        self.OST_XYZ.append([0, 0, 0])
        self.OST_LMN = []

        self.S_LMN = []
        self.LMN = []
        self.R_LMN = []
        self.N0 = []
        self.N1 = []
        self.WAV = 1.0
        self.G_LMN = []
        self.ORDER = []
        self.GRATING = []
        self.DISTANCE = []
        self.OP = []
        self.TOP_S = []
        self.TOP = 0
        self.ALPHA = [0.0]
        self.BULK_TRANS = []
        self.RP = []
        self.RS = []
        self.TP = []
        self.TS = []
        self.TTBE = []
        self.TT = 1.0
        return None

    def __EmptyCollect(self, pS, dC, WaveLength, j):
        """__EmptyCollect.

        Parameters
        ----------
        pS :
            pS
        dC :
            dC
        WaveLength :
            WaveLength
        j :
            j
        """
        Empty = np.asarray([])
        RayTraceType = 0
        ang = 0
        ValToSav = [Empty, Empty, pS, pS, Empty, Empty, dC, Empty, Empty, Empty, WaveLength, Empty, Empty, Empty, Empty, Empty, j, RayTraceType]
        self.__CollectData(ValToSav)

    def __WavePrecalc(self):
        """__WavePrecalc.
        """
        if (self.Wave != self.PreWave):
            self.N_Prec = []
            self.AlphaPrecal = []
            self.PreWave = self.Wave
            for i in range(0, self.n):
                (NP, AP) = n_wave_dispersion(self.SETUP, self.GlobGlass[i], self.Wave)
                self.N_Prec.append(NP)
                self.AlphaPrecal.append(AP)

    def __CollectData(self, ValToSav):
        """__CollectData.

        Parameters
        ----------
        ValToSav :
            ValToSav
        """
        [Glass, alpha, RayOrig, pTarget, HitObjSpace, LMNObjSpace, SurfNorm, ImpVec, ResVec, PrevN, CurrN, WaveLength, D, Ord, GrSpa, Name, j, RayTraceType] = ValToSav



        self.SURFACE.append(j)
        self.NAME.append(Name)
        self.GLASS.append(Glass)
        self.S_XYZ.append(RayOrig)
        self.T_XYZ.append(pTarget)
        self.XYZ[0] = self.S_XYZ[0]
        self.XYZ.append(pTarget)
        self.OST_XYZ.append(HitObjSpace)
        self.OST_LMN.append(LMNObjSpace)




        p = (np.asarray(RayOrig) - np.asarray(pTarget))
        dist = np.linalg.norm(p)
        self.DISTANCE.append(dist)
        self.OP.append((dist * PrevN))

        self.TOP = (self.TOP + (dist * PrevN))
        self.TOP_S.append(self.TOP)
        self.ALPHA.append(alpha)
        self.S_LMN.append(SurfNorm)
        self.LMN.append(ImpVec)
        self.R_LMN.append(ResVec)
        self.N0.append(PrevN)
        self.N1.append(CurrN)
        self.WAV = WaveLength
        self.G_LMN.append(D)
        self.ORDER.append(Ord)
        self.GRATING.append(GrSpa)
        if (self.val == 1):
            # Coating
            Rp2, Rs2, Tp2, Ts2, V  = self.CoatingFun(self.SDT[j].Coating, self.ang, self.Wave)
            (Rp, Rs, Tp, Ts) = FresnelEnergy(Glass, PrevN, CurrN, ImpVec, SurfNorm, ResVec, self.SETUP, self.Wave)
            if V == 1:
                Rp, Rs, Tp, Ts = Rp2, Rs2, Tp2, Ts2


        else:
            (Rp, Rs, Tp, Ts) = (0, 0, 0, 0)
        self.RP.append(Rp)
        self.RS.append(Rs)
        self.TP.append(Tp)
        self.TS.append(Ts)
        if ((RayTraceType == 0) or (RayTraceType == 1)):
            if (Glass == 'MIRROR'):
                self.tt = ((1.0 * (Rp + Rs)) / 2.0)
                self.BULK_TRANS.append(self.tt)
            if (Glass != 'MIRROR'):
                IT = np.exp(((- self.ALPHA[-2]) * dist))

                self.BULK_TRANS.append(IT)
                self.tt = (Tp + Ts) / 2.0
        else:
            self.tt = 1.0


        # si self.tt está vacio entonces es cero #
        if not self.tt:
            self.tt=0

        self.TTBE.append(self.tt*self.BULK_TRANS[-1])
        self.TT = (self.TT * self.tt*self.BULK_TRANS[-1])

        return None

    def CoatingFun(self, TRA, Theta, wav):
        R = np.asarray(TRA[0])
        A = np.asarray(TRA[1])
        W = np.asarray(TRA[2])
        THETA = np.asarray(TRA[3])

        T = 1-R-A

        if len(THETA) == 0:
            V = 0
            Rp, Rs, Tp, Ts = 0, 0, 0, 0
        else:
            V = 1
            idx_theta = (np.abs(THETA - Theta)).argmin()
            idx_w = (np.abs(W - wav)).argmin()

            T = T[idx_theta]
            R = R[idx_theta]
            A = A[idx_theta]


            T = T[idx_w]
            R = R[idx_w]
            A = A[idx_w]

            Rp, Rs, Tp, Ts = R, R, T, T

        return Rp, Rs, Tp, Ts, V

    def RestoreData(self):
        """RestoreData.
        """
        for ii in range(0, self.n):
            self.SDT[ii].RestoreSetup()
        self.SetData()

    def StoreData(self):
        """StoreData.
        """
        for ii in range(0, self.n):
            self.SDT[ii].SaveSetup()
        self.SetData()

    def SetData(self):
        """SetData.
        """
        self.SuTo = SUT(self.SDT)
        self.Object_Num = np.arange(0, self.n, 1)
        self.__SurFuncSuscrip()
        self.Pr3D.Prerequisites3SMath()
        self.INORM = InterNormalCalc(self.SDT, self.TypeTotal, self.Pr3D, self.HS)

    def SetSolid(self):
        """SetSolid.
        """
        self.__SurFuncSuscrip()
        self.Pr3D.Prerequisites3SMath()
        self.Pr3D.Prerequisites3DSolids()
        self.INORM = InterNormalCalc(self.SDT, self.TypeTotal, self.Pr3D, self.HS)
        self.AAA = self.Pr3D.AAA
        self.BBB = self.Pr3D.BBB
        self.DDD = self.Pr3D.DDD
        self.EEE = self.Pr3D.EEE
        self.TRANS_1A = self.Pr3D.TRANS_1A
        self.TRANS_2A = self.Pr3D.TRANS_2A

    def Parax(self, W):
        """Parax.

        Parameters
        ----------
        W :
            W
        """
        N_P = []
        for i in range(0, self.n):
            GlGl = self.GlobGlass[i]
            (NP, AP) = n_wave_dispersion(self.SETUP, GlGl, W)
            N_P.append(NP)
        Prx = ParaxCalc(N_P, self.SDT, self.SuTo, self.n, self.Glass)
        (self.SistemMatrix, self.S_Matrix, self.N_Matrix, self.a, self.b, self.c, self.d, self.EFFL, self.PPA, self.PPP, self.c_p, self.n_p, self.d_p) = Prx
        return Prx

    def TargSurf(self, tgsfP1):
        """TargSurf.

        Parameters
        ----------
        tgsfP1 :
            tgsfP1
        """
        tgsf = (tgsfP1 + 1)
        if (self.n >= tgsf > 0):
            self.Targ_Surf = tgsf
        else:
            self.Targ_Surf = self.n

    def SurfFlat(self, fltsf, Prep=0):
        """SurfFlat.

        Parameters
        ----------
        fltsf :
            fltsf
        Prep :
            Prep
        """
        if (self.n >= fltsf > 0):
            self.SuTo.Surface_Flattener = fltsf
        else:
            self.SuTo.Surface_Flattener = 0
        if (Prep != 0):
            self.Pr3D.Prerequisites3DSolids()

    def IgnoreVignetting(self, Prep=0):
        """IgnoreVignetting.

        Parameters
        ----------
        Prep :
            Prep
        """
        self.INORM.Disable_Inner = 0
        self.Pr3D.Disable_Inner = 0
        self.INORM.ExtraDiameter = 1
        self.Pr3D.ExtraDiameter = 1
        if (Prep != 0):
            self.Pr3D.Prerequisites3DSolids()

    def Vignetting(self, Prep=0):
        """Vignetting.

        Parameters
        ----------
        Prep :
            Prep
        """
        self.INORM.Disable_Inner = 1
        self.Pr3D.Disable_Inner = 1
        self.INORM.ExtraDiameter = 0
        self.Pr3D.ExtraDiameter = 0
        if (Prep != 0):
            self.Pr3D.Prerequisites3DSolids()


    def Trace(self, pS, dC, WaveLength):
        """Trace.

        Parameters
        ----------
        pS :
            pS
        dC :
            dC
        WaveLength :
            WaveLength
        """
        self.__CollectDataInit()
        ResVec = np.asarray(dC)
        RayOrig = np.asarray(pS)
        self.RAY = []
        self.Wave = WaveLength
        self.RAY.append(RayOrig)
        self.__WavePrecalc()
        j = 0
        Glass = self.GlobGlass[j]
        (PrevN, alpha) = (self.N_Prec[j], self.AlphaPrecal[j])
        j = 1
        SIGN = np.ones_like(ResVec)
        while True:
            if (j == self.Targ_Surf):
                break
            j_gg = j
            Glass = self.GlobGlass[j_gg]

            if (self.Glass[j] != 'NULL'):
                Proto_pTarget = (np.asarray(RayOrig) + ((np.asarray(ResVec) * 999999999.9) * SIGN))


                # start = timeit.default_timer()

                Output = self.INORM.InterNormal(RayOrig, Proto_pTarget, j, j)

                # stop = timeit.default_timer()
                # execution_time = stop - start
                # self.ExectTime.append(execution_time)








                (SurfHit, SurfNorm, pTarget, GooveVect, HitObjSpace, LMNObjSpace, j) = Output

                if (SurfHit == 0):
                    break

                ImpVec = np.asarray(ResVec)
                (CurrN, alpha) = (self.N_Prec[j_gg], self.AlphaPrecal[j_gg])
                S = ImpVec
                R = SurfNorm
                N = PrevN
                Np = CurrN
                D = GooveVect

                Ord = self.SDT[j].Diff_Ord
                # print(Ord)

                GrSpa = self.SDT[j].Grating_D
                Secuent = 0

                (ResVec, CurrN, sign ,self.ang) = self.SDT[j].PHYSICS.calculate(S, R, N, Np, D, Ord, GrSpa, self.Wave, Secuent)


                SIGN = (SIGN * sign)
                Name = self.SDT[j].Name
                RayTraceType = 0
                ValToSav = [Glass, alpha, RayOrig, pTarget, HitObjSpace, LMNObjSpace, SurfNorm, ImpVec, ResVec, PrevN, CurrN, WaveLength, D, Ord, GrSpa, Name, j, RayTraceType]
                self.__CollectData(ValToSav)
                PrevN = CurrN
                RayOrig = pTarget
                self.RAY.append(RayOrig)

            if self.Glass[j] == 'NULL':
                ang = 0

                ValToSav = [Glass, alpha, RayOrig, pTarget, HitObjSpace, LMNObjSpace, SurfNorm, ImpVec, ResVec, PrevN, CurrN, WaveLength, D, Ord, GrSpa, Name, j, RayTraceType]
                self.__CollectData(ValToSav)

            if self.Glass[j] == 'ABSORB':
                break








            j = (j + 1)

        if (len(self.GLASS) == 0):
            self.__CollectDataInit()
            self.val = 0
            self.__EmptyCollect(RayOrig, ResVec, WaveLength, j)

        self.ray_SurfHits = np.asarray(self.RAY)

        AT = np.transpose(self.ray_SurfHits)
        self.Hit_x = AT[0]
        self.Hit_y = AT[1]
        self.Hit_z = AT[2]



        # execution_time=np.mean(np.asarray(self.ExectTime))
        # print(str(execution_time)," Segmentos")


        self.ExectTime=[]

    def NsTrace(self, pS, dC, WaveLength):
        """NsTrace.

        Parameters
        ----------
        pS :
            pS
        dC :
            dC
        WaveLength :
            WaveLength
        """
        global j_gg
        pS = np.asarray(pS)
        count = 0
        a = 1
        b = 2
        self.__CollectDataInit()
        ResVec = dC
        RayOrig = pS
        self.RAY = []
        self.Wave = WaveLength
        self.RAY.append(RayOrig)
        self.__WavePrecalc()
        j = 0
        Glass = self.GlobGlass[j]
        (PrevN, alpha) = (self.N_Prec[j], self.AlphaPrecal[j])
        j = 0
        SIGN = 1
        while True:
            if (j == self.Targ_Surf):
                break
            (a, b, c, PreSurfHit) = self.__NonSequentialChooser(SIGN, RayOrig, ResVec, j)
            if (PreSurfHit == 0):
                break
            if (a < b):
                j_gg = b
            if (a > b):
                j_gg = (b - 1)
                if (self.Glass[(b - 1)] == 'MIRROR'):
                    j_gg = b
                if (self.Glass[b] == 'MIRROR'):
                    j_gg = b
            if (a == b):
                j_gg = (b - 1)
            j = b
            jj = b
            j = self.GlassOnSide[j]
            j_gg = self.GlassOnSide[j_gg]
            Glass = self.GlobGlass[j_gg]
            if ((j == 0) or (count > self.NsLimit) or (a == self.n)):
                break
            if (self.Glass[j] != 'NULL'):
                Proto_pTarget = (np.asarray(RayOrig) + ((np.asarray(ResVec) * 999999999.9) * SIGN))
                Output = self.INORM.InterNormal(RayOrig, Proto_pTarget, j, jj)
                (SurfHit, SurfNorm, pTarget, GooveVect, HitObjSpace, LMNObjSpace, j) = Output
                if (SurfHit == 0):
                    break
                ImpVec = np.asarray(ResVec)
                (CurrN, alpha) = (self.N_Prec[j_gg], self.AlphaPrecal[j_gg])
                S = np.asarray(ImpVec)
                R = np.asarray(SurfNorm)
                N = PrevN
                Np = CurrN
                if ((self.SDT[j].Solid_3d_stl == 'None') and (self.TypeTotal[jj] == 1)):
                    if (N == 1):
                        Np = CurrN
                    else:
                        N = CurrN
                        Np = 1
                D = GooveVect

                Ord = self.SDT[j].Diff_Ord
                # print(Ord)
                # Ord = self.SDT[j].Diff_Ord
                GrSpa = self.SDT[j].Grating_D
                Secuent = 0
                ResVec_N, R_N, N_N, Np_N = ResVec, R, N, Np
                (ResVec, CurrN, sign ,ang) = self.SDT[j].PHYSICS.calculate(ResVec_N, R_N, N_N, Np_N, D, Ord, GrSpa, self.Wave, Secuent)

                # Coating
                (Rp0, Rs0, Tp0, Ts0) = FresnelEnergy(self.Glass[j], N, Np, ResVec, R, ResVec, self.SETUP, self.Wave)
                Rp2, Rs2, Tp2, Ts2, V = self.CoatingFun(self.SDT[j].Coating, self.ang, self.Wave)
                if V == 1:
                    Rp0, Rs0, Tp0, Ts0 = Rp2, Rs2, Tp2, Ts2

                    # print(V, Rp0, Rs0, Tp0, Ts0, j)


                self.tt = 1.0
                if (self.Glass[j] != 'MIRROR'):
                    self.tt = ((Tp0 + Ts0) / 2.0)


                if self.energy_probability==1:
                    PROB = prob(self.tt)[0]
                    # print(PROB)

                    if (PROB > 0):
                        Secuent = 1
                        (ResVec, CurrN, sign ,ang) = self.SDT[j].PHYSICS.calculate(ResVec_N, R_N, N_N, Np_N, D, Ord, GrSpa, self.Wave, Secuent)

                SIGN = (SIGN * sign)

                Name = self.SDT[j].Name
                RayTraceType = 1
                ValToSav = [Glass, alpha, RayOrig, pTarget, HitObjSpace,LMNObjSpace, SurfNorm, ImpVec, ResVec, PrevN, CurrN, WaveLength, D, Ord, GrSpa, Name, j, RayTraceType]
                self.__CollectData(ValToSav)
                if (a == b):
                    PrevN = PrevN
                else:
                    PrevN = CurrN
                RayOrig = pTarget
                self.RAY.append(RayOrig)






            if self.Glass[j] == 'NULL':
                ang = 0
                ValToSav = [Glass, alpha, RayOrig, pTarget, HitObjSpace, SurfNorm, ImpVec, ResVec, PrevN, CurrN, WaveLength, D, Ord, GrSpa, Name, j, RayTraceType, ang]
                self.__CollectData(ValToSav)

            if self.Glass[j] == 'ABSORB':
                break






            count = (count + 1)
        if (len(self.GLASS) == 0):
            self.__CollectDataInit()
            self.val = 0
            self.__EmptyCollect(pS, dC, WaveLength, j)
        self.ray_SurfHits = np.asarray(self.RAY)
        AT = np.transpose(self.ray_SurfHits)
        self.Hit_x = AT[0]
        self.Hit_y = AT[1]
        self.Hit_z = AT[2]

    def FastTrace(self, pS, dC, WaveLength):
        """Trace.

        Parameters
        ----------
        pS :
            pS
        dC :
            dC
        WaveLength :
            WaveLength
        """
        self.__CollectDataInit()
        ResVec = np.asarray(dC)
        RayOrig = np.asarray(pS)
        self.RAY = []
        self.Wave = WaveLength
        self.RAY.append(RayOrig)
        self.__WavePrecalc()
        j = 0
        Glass = self.GlobGlass[j]
        PrevN = self.N_Prec[j]
        j = 1
        SIGN = np.ones_like(ResVec)

        while True:
            if (j == self.Targ_Surf):
                break

            j_gg = j
            Glass = self.GlobGlass[j_gg]

            if ((self.Glass[j] != 'NULL') and (self.Glass[j] != 'ABSORB')):
                Proto_pTarget = (np.asarray(RayOrig) + ((ResVec * 999999999.9) * SIGN))

                Output = self.INORM.InterNormalFast(RayOrig, Proto_pTarget, j)

                (SurfHit, SurfNorm, pTarget, GooveVect, HitObjSpace, j) = Output

                if (SurfHit != 0):

                    ImpVec = ResVec
                    CurrN = self.N_Prec[j_gg]
                    S = ImpVec
                    R = SurfNorm
                    N = PrevN
                    Np = CurrN
                    D = GooveVect

                    Ord = self.SDT[j].Diff_Ord
                    GrSpa = self.SDT[j].Grating_D

                    (ResVec, CurrN, sign ,ang) = self.SDT[j].PHYSICS.calculate(S, R, N, Np, D, Ord, GrSpa, self.Wave, 0)

                    SIGN = (SIGN * sign)

                    PrevN = CurrN
                    RayOrig = pTarget
                    self.RAY.append(RayOrig)

                else:
                    break

            j = (j + 1)



        self.ray_SurfHits = np.asarray(self.RAY)


        self.ExectTime=[]



# from numba import jit
# @jit(forceobj=True)

import numpy as np
from .SurfTools import surface_tools as SUT
# from scipy.optimize import fsolve
# import math

class Hit_Solver():
    """Hit_Solver.
    """


    def __init__(self, SurfData):
        """__init__.

        Parameters
        ----------
        SurfData :
            SurfData
        """
        self.SDT = SurfData
        self.n = len(self.SDT)
        self.SuTo = SUT(SurfData)
        self.vj = 0
        self.vevaX = 0
        self.vevaY = 0
        self.NP_x1 = 0
        self.NP_y1 = 0
        self.NP_z1 = 0
        self.MN = 0
        self.LN = 0
        self.SuTo.ErrSurfCase = 0
        self.delta = 0.04
        self.delta2 = 2.0 * self.delta
        self.deltaLineCurve = 1e-13
        self.Parray=np.asarray([0.0, 0.0, 0.0])

        self.vx = np.asarray([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.vy = np.asarray([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.vz = np.asarray([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        self.vxi = [self.delta, 0.0, 0.0, - self.delta, 0.0, 0.0]
        self.vyi = [0.0, self.delta, 0.0, 0.0, - self.delta, 0.0]
        self.vzi = [0.0, 0.0, self.delta, 0.0, 0.0, - self.delta]



    # def SurfDer(self, x, y, z):
    #     """SurfDer.

    #     Parameters
    #     ----------
    #     x :
    #         x
    #     y :
    #         y
    #     z :
    #         z
    #     """

    #     # self.vx = np.asarray([(x + self.delta), x, x, (x - self.delta), x, x])
    #     # self.vy = np.asarray([y, (y + self.delta), y, y, (y - self.delta), y])
    #     # self.vz = np.asarray([z, z, (z + self.delta), z, z, (z - self.delta)])


    #     self.vx1 = np.asarray([(x + self.delta), x, x, (x - self.delta), x, x, (x + self.delta2), x, x, (x - self.delta2), x, x])
    #     self.vy1 = np.asarray([y, (y + self.delta), y, y, (y - self.delta), y, y, (y + self.delta2), y, y, (y - self.delta2), y])
    #     self.vz1 = np.asarray([z, z, (z + self.delta), z, z, (z - self.delta), z, z, (z + self.delta2), z, z, (z - self.delta2)])


    #     F1 = self.SuTo.SurfaceShape(self.vx1, self.vy1, self.vj) - self.vz1


    #     V = F1 / (2.0 * self.delta)


    #     Dx = V[0] - V[3]
    #     Dy = V[1] - V[4]
    #     Dz = V[2] - V[5]

    #     Dr = np.sqrt((((Dx ** 2.) + (Dy ** 2.)) + (Dz ** 2.)))

    #     return ((Dx / Dr), (Dy / Dr), (Dz / Dr))

    def SurfDer(self, x, y, z):
        #http://www2.math.umd.edu/~dlevy/classes/amsc466/lecture-notes/differentiation-chap.pdf
        """SurfDer.

        Parameters
        ----------
        x :
            x
        y :
            y
        z :
            z
        """

        self.vx1 = np.asarray([(x + self.delta), x, x, (x - self.delta), x, x, (x + self.delta2), x, x, (x - self.delta2), x, x])
        self.vy1 = np.asarray([y, (y + self.delta), y, y, (y - self.delta), y, y, (y + self.delta2), y, y, (y - self.delta2), y])
        self.vz1 = np.asarray([z, z, (z + self.delta), z, z, (z - self.delta), z, z, (z + self.delta2), z, z, (z - self.delta2)])


        F = self.SuTo.SurfaceShape(self.vx1, self.vy1, self.vj) - self.vz1

        AX1=-F[6]
        AX2=8.*F[0]
        AX3=-8.*F[3]
        AX4=F[9]
        Dx=(AX1+AX2+AX3+AX4)/(12.0*self.delta)

        AY1=-F[6+1]
        AY2=8.*F[0+1]
        AY3=-8.*F[3+1]
        AY4=F[9+1]
        Dy=(AY1+AY2+AY3+AY4)/(12.0*self.delta)

        AZ1=-F[6+2]
        AZ2=8.*F[0+2]
        AZ3=-8.*F[3+2]
        AZ4=F[9+2]
        Dz=(AZ1+AZ2+AZ3+AZ4)/(12.0*self.delta)

        Dr = np.sqrt((((Dx ** 2.) + (Dy ** 2.)) + (Dz ** 2.)))
        return ((Dx / Dr), (Dy / Dr), (Dz / Dr))








    def __surface_Derivative(self,x,y,z,j):
        #http://www2.math.umd.edu/~dlevy/classes/amsc466/lecture-notes/differentiation-chap.pdf
        delta=0.000001
        # more precision but slower execution
        AX1=-(self.__xyzF(x+(2*delta),y,z,j))
        AX2=8.0*(self.__xyzF(x+delta,y,z,j))
        AX3=-8.0*(self.__xyzF(x-delta,y,z,j))
        AX4=(self.__xyzF(x-(2*delta),y,z,j) )
        Dx=(AX1+AX2+AX3+AX4)/(12*delta)

        AY1=-(self.__xyzF(x,y+2*delta,z,j))
        AY2=8.0*(self.__xyzF(x, y+delta,z,j))
        AY3=-8.0*(self.__xyzF(x,y-delta,z,j))
        AY4=(self.__xyzF(x, y-2*delta,z,j)  )
        Dy=(AY1+AY2+AY3+AY4)/(12*delta)

        AZ1=-(self.__xyzF(x,y,z+2*delta,j))
        AZ2=8.0*(self.__xyzF(x,y, z+delta,j))
        AZ3=-8.0*(self.__xyzF(x,y,z-delta,j))
        AZ4=(self.__xyzF(x, y, z-2*delta,j) )
        Dz=(AZ1+AZ2+AZ3+AZ4)/(12*delta)

        # Dx=(self.__xyzF(x+delta,y,z,j)-self.__xyzF(x-delta,y,z,j))/(2.0*delta)
        # Dy=(self.__xyzF(x,y+delta,z,j)-self.__xyzF(x,y-delta,z,j))/(2.0*delta)
        # Dz=(self.__xyzF(x,y,z+delta,j)-self.__xyzF(x,y,z-delta,j))/(2.0*delta)

        Dr=np.sqrt((Dx*Dx)+(Dy*Dy)+(Dz*Dz))
        return Dx/Dr,Dy/Dr,Dz/Dr








    def __DerLineCurve(self, gf):
        """__DerLineCurve.

        Parameters
        ----------
        gf :
            gf
        """
        self.Parray[0]= gf + self.deltaLineCurve
        self.Parray[1]= gf
        self.Parray[2]= gf - self.deltaLineCurve

        kl=self.Parray - self.NP_z1
        X = (( kl * self.LN ) + self.NP_x1)
        Y = (( kl * self.MN ) + self.NP_y1)


        SP = self.SuTo.SurfaceShape(X, Y, self.vj)
        AA = (SP - self.Parray)

        Dz = ((AA[0] - AA[2]) / (2.0 * self.deltaLineCurve))
        return Dz, AA[1]

    def SolveHit(self, Px1, Py1, Pz1, L, M, N, j):
        """SolveHit.

        Parameters
        ----------
        Px1 :
            Px1
        Py1 :
            Py1
        Pz1 :
            Pz1
        L :
            L
        M :
            M
        N :
            N
        j :
            j
        """

        # if (len(self.SDT[j].Error_map) == 0):
        #     self.SuTo.ErrSurfCase = 0
        # else:
        #     self.SuTo.ErrSurfCase = 0

        self.vevaX = 0.
        self.vevaY = 0.
        self.vj = j

        # cnt = 0
        PP_z2 = Pz1
        P_z2 = Pz1
        self.NP_x1 = Px1
        self.NP_y1 = Py1
        self.NP_z1 = Pz1
        self.MN = (M / N)
        self.LN = (L / N)

        for i in range (0,30):

            DerFdeX, FdeX = self.__DerLineCurve(PP_z2)

            PP_z2 = (PP_z2 - (FdeX / DerFdeX))

            if (np.abs((PP_z2 - P_z2)) > 1e-9):
                P_z2 = PP_z2
            else:
                break



        P_z2 = PP_z2
        P_x2 = (((P_z2 - self.NP_z1) * self.LN) + self.NP_x1)
        P_y2 = (((P_z2 - self.NP_z1) * self.MN) + self.NP_y1)

        # self.SuTo.ErrSurfCase = 0


        if (len(self.SDT[j].Error_map) != 0):
            PP_z2 = P_z2
            cnt = 0
            while True:
                DerFdeX , FdeX= self.__DerLineCurve(PP_z2)
                PP_z2  = (PP_z2 - (FdeX / DerFdeX))
                if (np.abs((PP_z2 - P_z2)) < 1e-9):
                    break
                else:
                    P_z2 = PP_z2
                if (cnt == 3):
                    break
                cnt += 1
            P_z2 = PP_z2
            P_x2 = (((P_z2 - self.NP_z1) * self.LN) + self.NP_x1)
            P_y2 = (((P_z2 - self.NP_z1) * self.MN) + self.NP_y1)

        self.vevaX = P_x2
        self.vevaY = P_y2


        return (P_x2, P_y2, P_z2)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Multicore"""

import multiprocessing
import time
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

start_time = time.time()

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0.0
P_Obj.Thickness = 10
P_Obj.Glass = "AIR"
P_Obj.Diameter = 30.0

# ______________________________________#

L1a = Kos.surf()
L1a.Rc = 9.284706570002484E+001
L1a.Thickness = 6.0
L1a.Glass = "BK7"
L1a.Diameter = 30.0
L1a.Axicon = 0

# ______________________________________#

L1b = Kos.surf()
L1b.Rc = -3.071608670000159E+001
L1b.Thickness = 3.0
L1b.Glass = "F2"
L1b.Diameter = 30

# ______________________________________#

L1c = Kos.surf()
L1c.Rc = -7.819730726078505E+001
L1c.Thickness = 9.737604742910693E+001
L1c.Glass = "AIR"
L1c.Diameter = 30

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0.0
P_Ima.Thickness = 0.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 3.0
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, L1a, L1b, L1c, P_Ima]
config_1 = Kos.Setup()

# ______________________________________#

Doblete1 = Kos.system(A, config_1)


# ______________________________________#

def trax1(xyz, lmn, w, q):
    Rayos = Kos.raykeeper(Doblete1)
    start_time = time.time()
    for i in range(0, 700):
        Doblete1.Trace(xyz, lmn, w)
        Rayos.push()
    A = Rayos.pick(-1)
    Rayos.clean()
    q.put(A[0])
    print("--- %s seconds ---" % (time.time() - start_time))


# ______________________________________#

if __name__ == '__main__':
    start_time = time.time()
    pSource_0 = [1, 0, 0.0]
    dCos = [0.0, np.sin(np.deg2rad(0)), np.cos(np.deg2rad(0))]
    w = 0.5
    q = multiprocessing.Queue()
    p1 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.1, q))
    p2 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.2, q))
    p3 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.3, q))
    p4 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.4, q))
    p5 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.5, q))
    p6 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.6, q))
    p7 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.7, q))
    p8 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.8, q))
    p9 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.1, q))
    p10 = multiprocessing.Process(target=trax1, args=(pSource_0, dCos, w + 0.2, q))

    p1.start()
    p2.start()
    p3.start()
    p4.start()
    p5.start()
    p6.start()
    p7.start()
    p8.start()
    p9.start()
    p10.start()

    p1.join()
    p2.join()
    p3.join()
    p4.join()
    p5.join()
    p6.join()
    p7.join()
    p8.join()
    p9.join()
    p10.join()

    print(".................................................")
    print("Total time :")
    print("--- %s seconds ---" % (time.time() - start_time))

    for i in range(0, 10):
        A = q.get()
        print(len(A))



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp-Tel_2M_Atmospheric_Refraction_Corrector"""

import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
import numpy as np
import matplotlib.pyplot as plt
import scipy





def MyPlot(RK,surf, figure= "Spot", mk=["x"], col = [[0.8,0.0,0.0]]):
    fig = plt.figure(figure)
    ax = fig.add_subplot()
    i=0
    for rk in RK:
        x,y,z,l,m,n = rk.pick(surf[i], coordinates="local")
        plt.scatter(x, y,marker = mk[i], color = col[i])
        i = i + 1
    # https://matplotlib.org/3.1.1/api/markers_api.html
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlabel('x (mm)')
    ax.set_ylabel('y (mm)')
    ax.set_title('Spot diagram')
    plt.show()



def R_RMS_delta(Z1, L, M, N, X0, Y0):
    X1 = ((L / N) * Z1) + X0
    Y1 = ((M / N) * Z1) + Y0
    cenX = np.mean(X1)
    cenY = np.mean(Y1)
    x1 = (X1 - cenX)
    y1 = (Y1 - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def BestFocus(X, Y, Z, L, M, N, system, mod=1):
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    if mod ==1:
        system.SDT[-2].Thickness = system.SDT[-2].Thickness + v[0]
        system.SetData()
        system.SetSolid()
    return system, v[0]


def BestRMS(system,raykeeper):
    x,y,z,l,m,n = raykeeper.pick(-1 , coordinates="local")
    system, deltaZ = BestFocus(x, y, z, l, m, n, system, mod = 0)
    rms = R_RMS_delta(deltaZ,l,m,n,x,y)
    return rms
# _________________________________________________________________#

T = 0 # Prism rotation
ZenitDist = 55.0


A = 1.55/2
P_Obj = Kos.surf(Thickness = 4452.2, Glass = "AIR", Diameter = 2118.0)
P_Obj.Drawing = 0

Thickness = 3452.2
M1 = Kos.surf()
M1.Rc = -9638.0
M1.Thickness = -Thickness
M1.k = -1.07731
M1.Glass = "MIRROR"
M1.Diameter = 1059.0 * 2.0
M1.InDiameter = 250 * 2.0

M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness
M2.k = -4.3281
M2.Glass = "MIRROR"
M2.Diameter = 336.5 * 2.0

M1Vertex = Kos.surf()
M1Vertex.Thickness = 737.525
M1Vertex.Diameter = 200
M1Vertex.TiltZ=T
M1Vertex.AxisMove = 1
M1Vertex.Drawing =0


C1 = Kos.surf()
C1.Thickness = 5
C1.Glass = "BK7"
C1.Diameter = 100

C2 = Kos.surf()
C2.Thickness = C1.Thickness
C2.Glass = "F2"
C2.Diameter = 100
C2.TiltY = A
C2.AxisMove = 0

C3 = Kos.surf()
C3.Thickness = C1.Thickness
C3.Glass = "AIR"
C3.Diameter = 100
C3.TiltZ = -2 * M1Vertex.TiltZ
C3.AxisMove = 1


C4 = Kos.surf()
C4.Thickness = C1.Thickness
C4.Glass = "BK7"
C4.Diameter = 100

C5 = Kos.surf()
C5.Thickness = C1.Thickness
C5.Glass = "F2"
C5.Diameter = 100
C5.TiltY = A
C5.AxisMove = 0

C6 = Kos.surf()
C6.Thickness = C1.Thickness + 277.252
C6.Glass = "AIR"
C6.Diameter = 100

RZ = Kos.surf(Diameter = 100)
RZ.TiltZ =  M1Vertex.TiltZ
RZ.AxisMove = 1

P_Ima = Kos.surf(Diameter = 100.0)

A = [P_Obj, M1, M2, M1Vertex, C1, C2, C3, C4, C5, C6, RZ, P_Ima]
Config_1 = Kos.Setup()
Tel = Kos.system(A, Config_1)

# _________________________________________________________________#

Rays1 = Kos.raykeeper(Tel)
Rays2 = Kos.raykeeper(Tel)
Rays3 = Kos.raykeeper(Tel)





W = 0.60169
sup = 1  # Difining M1 as enter pupil diameter
AperVal = 2000
AperType = "EPD"  # "STOP"
Pup = Kos.PupilCalc(Tel, sup, W, AperType, AperVal)
Pup.Samp = 11
Pup.FieldType = "angle"

Pup.AtmosRef = 1
Pup.T = 283.15  # k
Pup.P = 101300  # Pa
Pup.H = 0.5  # Humidity ratio 1 to 0
Pup.xc = 400  # ppm
Pup.lat = 31  # degrees
Pup.h = 2800  # meters
Pup.l1 = W  # micron
Pup.l2 = 0.50169  # micron
Pup.z0 = ZenitDist  # degrees
Pup.Ptype = "hexapolar"
Pup.FieldX = 0.0


W = [0.50169, 0.60169, 0.70169]
RAYS=[Rays1, Rays2, Rays3]
i = 0
for w in W:
    Pup.l2 = w
    x, y, z, L, M, N = Pup.Pattern2Field()
    Kos.TraceLoop(x, y, z, L, M, N, w, RAYS[i], clean = 1)
    i = i + 1

MK = [".", ".", "."]
COL = [[0.8,0.0,0.0], [0.0,0.8,0.0], [0.0,0.0,0.8]]
SURF = [-1, -1, -1]
MyPlot(RAYS, SURF, figure= "Spot", mk = MK, col = COL)




import numpy as np
import KrakenOS as Kos
from .PupilTool import PupilCalc, SolveVectCross

class Seidel():
    """Seidel.
    """


    def __init__(self, PUPIL):
        """__init__.

        Parameters
        ----------
        PUPIL :
            PUPIL
        """
        self.PUPIL = PUPIL
        self.SYSTEM = self.PUPIL.SYSTEM
        self.calculate()

    def calculate(self):
        """calculate.
        """
        sup = self.PUPIL.Surf
        W = self.PUPIL.W
        ApType = self.PUPIL.ApertureType
        ApVal = self.PUPIL.ApertureValue
        fx = self.PUPIL.FieldX
        fy = self.PUPIL.FieldY
        field = np.sqrt(((fx ** 2) + (fy ** 2)))
        fieldType = self.PUPIL.FieldType
        Pup = PupilCalc(self.SYSTEM, sup, W, ApType, ApVal)
        Prx = self.SYSTEM.Parax(W)
        (SistemMatrix, S_Matrix, N_Matrix, a, b, c, d, EFFL, PPA, PPP, C, NN, d) = Prx
        n_1 = np.asarray(NN)
        n_sign = np.sign(NN)
        n_abs = np.abs(NN)
        for si in range(0, len(n_sign)):
            n_abs[si:] = (n_sign[si] * n_abs[si:])
        n_1 = np.copy(n_abs)
        n_2 = np.copy(n_abs)
        n_1 = np.insert(n_1, 0, n_1[0])
        n_elements = self.SYSTEM.n
        c = []
        d = []
        kon = []
        asp2 = []
        asp4 = []
        asp6 = []
        d.append(0)
        for i in range(0, n_elements):
            rc = self.SYSTEM.SDT[i].Rc
            if (rc == 0):
                c.append(rc)
            else:
                c.append((1 / rc))
            t = self.SYSTEM.SDT[i].Thickness
            d.append(t)
            kon.append(self.SYSTEM.SDT[i].k)
            asp2.append(self.SYSTEM.SDT[i].AspherData[0])
            asp4.append(self.SYSTEM.SDT[i].AspherData[1])
            asp6.append(self.SYSTEM.SDT[i].AspherData[2])
        c.append(0)
        d.append(0)
        kon.append(0)
        asp2.append(0)
        asp4.append(0)
        asp6.append(0)
        t = d[1]
        d[1] = (d[1] - Pup.PosPupInp[2])
        Diam = (Pup.RadPupInp * 2.0)
        c = np.asarray(c)
        epsilon = np.asarray(kon)
        asp2 = np.asarray(asp2)
        asp4 = np.asarray(asp4)
        asp6 = np.asarray(asp6)
        c_orig = np.copy(c)
        c = (c + (2.0 * asp2))
        u = []
        h = []
        u_bar = []
        h_bar = []
        E = []
        Delta_unn = []
        Delta_nn = []
        Delta_n = []
        Delta_1sn = []
        A = []
        A_bar = []
        sI = []
        sII = []
        sIII = []
        sIV = []
        sV = []
        sI_k = []
        sII_k = []
        sIII_k = []
        sIV_k = []
        sV_k = []
        if (fieldType == 'angle'):
            Theta_marg = 0
            Theta_bar = (- np.deg2rad(field))
        PosObj = (t + d[1])
        if (fieldType == 'height'):
            Theta_marg = np.arctan2(Pup.RadPupInp, PosObj)
            Theta_bar = (- np.arctan2(field, PosObj))
        u.append(Theta_marg)
        h.append((Diam / 2.0))
        u_bar.append(Theta_bar)
        h_bar.append(0)
        E.append(0)
        for k in range(0, (len(c) - 1)):
            nk = n_1[k]
            nkm1 = n_2[k]
            u.append((((nk * u[k]) - ((h[k] * c[k]) * (nkm1 - nk))) / nkm1))
            h.append((h[k] + (u[(k + 1)] * d[(k + 1)])))
            u_bar.append((((nk * u_bar[k]) - ((h_bar[k] * c[k]) * (nkm1 - nk))) / nkm1))
            h_bar.append((h_bar[k] + (u_bar[(k + 1)] * d[(k + 1)])))
            Epro = (((- d[(k + 1)]) / ((nkm1 * h[(k + 1)]) * h[k])) + E[k])
            E.append(Epro)
            Delta_unn.append(((u[(k + 1)] / nkm1) - (u[k] / nk)))
            Delta_nn.append(((1.0 / nkm1) - (1.0 / nk)))
            Delta_n.append((nkm1 - nk))
            Delta_1sn.append(((1.0 / (nkm1 ** 2)) - (1 / (nk ** 2))))
            A.append((nk * ((h[k] * c[k]) + u[k])))
        Abar = (1.0 * ((h_bar[0] * c[0]) + u_bar[0]))
        H = ((A[0] * h_bar[0]) - (Abar * h[0]))
        for k in range(0, (len(c) - 1)):
            A_bar.append(((H / h[k]) * (((A[k] * h[k]) * E[k]) - 1)))
            sI.append((((A[k] ** 2.0) * h[k]) * Delta_unn[k]))
            sII.append((((A[k] * A_bar[k]) * h[k]) * Delta_unn[k]))
            sIII.append((((A_bar[k] ** 2.0) * h[k]) * Delta_unn[k]))
            sIV.append((((H ** 2.0) * c[k]) * Delta_nn[k]))
            if (A[k] != 0):
                P = (c[k] * Delta_nn[k])
                p1 = ((H ** 2.0) * P)
                p2 = (((A_bar[k] ** 2.0) * h[k]) * Delta_unn[k])
                sV.append(((A_bar[k] / A[k]) * (p1 + p2)))
            else:
                sV.append(0)
            a = ((epsilon[k] * (c[k] ** 3.0)) * (h[k] ** 4.0))
            sI_k.append(((a * Delta_n[k]) * ((h_bar[k] / h[k]) ** 0)))
            sII_k.append(((a * Delta_n[k]) * ((h_bar[k] / h[k]) ** 1)))
            sIII_k.append(((a * Delta_n[k]) * ((h_bar[k] / h[k]) ** 2)))
            sIV_k.append((a * 0.0))
            sV_k.append(((a * Delta_n[k]) * ((h_bar[k] / h[k]) ** 3)))
        sI_a = []
        sII_a = []
        sIII_a = []
        sIV_a = []
        sV_a = []
        c = c_orig
        Asp = (asp4 - ((asp2 / 4.0) * (((4.0 * (asp2 ** 2)) + ((6.0 * asp2) * c)) + (3.0 * (c ** 2.0)))))
        for k in range(0, (len(c) - 1)):
            a = (((8.0 * Asp[k]) * (h[k] ** 4.0)) * Delta_n[k])
            sI_a.append((a * ((h_bar[k] / h[k]) ** 0)))
            sII_a.append((a * ((h_bar[k] / h[k]) ** 1)))
            sIII_a.append((a * ((h_bar[k] / h[k]) ** 2)))
            sIV_a.append((a * 0.0))
            sV_a.append((a * ((h_bar[k] / h[k]) ** 3)))
        sI = np.asarray(sI)
        sII = np.asarray(sII)
        sIII = np.asarray(sIII)
        sIV = np.asarray(sIV)
        sV = np.asarray(sV)
        sI_k = np.asarray(sI_k)
        sII_k = np.asarray(sII_k)
        sIII_k = np.asarray(sIII_k)
        sIV_k = np.asarray(sIV_k)
        sV_k = np.asarray(sV_k)
        sI_a = np.asarray(sI_a)
        sII_a = np.asarray(sII_a)
        sIII_a = np.asarray(sIII_a)

        self.si = ((sI - sI_k) - sI_a)
        self.sii = ((sII - sII_k) - sII_a)
        self.siii = ((sIII - sIII_k) - sIII_a)
        self.siv = ((sIV - sIV_k) - sIV_a)
        self.sv = ((sV - sV_k) - sV_a)
        #---------------------------------#
        self.SAC = [self.si, self.sii, self.siii, self.siv, self.sv]
        self.SAC_TOTAL = [np.sum(self.si), np.sum(self.sii), np.sum(self.siii), np.sum(self.siv), np.sum(self.sv)]
        self.SAC_NM = ['si', 'sii', 'siii', 'siv', 'sv']

        Units = 1000.0
        self.W040 = (((Units * (1 / 8.0)) * self.si) / W)
        self.W131 = (((Units * (1 / 2.0)) * self.sii) / W)
        self.W222 = (((Units * (1 / 2.0)) * self.siii) / W)
        self.W220 = (((Units * (1 / 4.0)) * self.siv) / W)
        self.W311 = (((Units * (1 / 2.0)) * self.sv) / W)
        #---------------------------------#
        self.SCW = [self.W040, self.W131, self.W222, self.W220, self.W311]
        self.SCW_TOTAL = [np.sum(self.W040), np.sum(self.W131), np.sum(self.W222), np.sum(self.W220), np.sum(self.W311)]
        self.SCW_NM = ['W040', 'W131', 'W222', 'W220', 'W311']

        self.TSPH = (self.si / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TSCO = (self.sii / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TTCO = ((3.0 * self.sii) / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TAST = ((- self.siii) / (u[(- 1)] * n_1[(- 1)]))
        self.TPFC = ((- self.siv) / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TSFC = ((self.siii + self.siv) / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TTFC = (((3.0 * self.siii) + self.siv) / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        self.TDIS = ((- self.sv) / ((2.0 * u[(- 1)]) * n_1[(- 1)]))
        #---------------------------------#

        self.TAC = [self.TSPH, self.TSCO, self.TTCO, self.TAST, self.TPFC, self.TSFC, self.TTFC, self.TDIS]
        self.TAC_TOTAL = [np.sum(self.TSPH), np.sum(self.TSCO), np.sum(self.TTCO), np.sum(self.TAST), np.sum(self.TPFC), np.sum(self.TSFC), np.sum(self.TTFC), np.sum(self.TDIS)]
        self.TAC_NM = ['TSPH', 'TSCO', 'TTCO', 'TAST', 'TPFC', 'TSFC', 'TTFC', 'TDIS']

        #---------------------------------#
        self.LSPH = (self.si / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LSCO = (self.sii / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LTCO = ((3.0 * self.sii) / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LAST = (self.siii / ((u[(- 1)] * u[(- 1)]) * n_1[(- 1)]))
        self.LPFC = (self.siv / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LSFC = ((self.siii + self.siv) / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LTFC = (((3.0 * self.siii) + self.siv) / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        self.LDIS = ((- self.sv) / (((2.0 * u[(- 1)]) * u[(- 1)]) * n_1[(- 1)]))
        #---------------------------------#
        self.LAC = [self.LSPH, self.LSCO, self.LTCO, self.LAST, self.LPFC, self.LSFC, self.LTFC, self.LDIS]
        self.LAC_TOTAL = [np.sum(self.LSPH), np.sum(self.LSCO), np.sum(self.LTCO), np.sum(self.LAST), np.sum(self.LPFC), np.sum(self.LSFC), np.sum(self.LTFC), np.sum(self.LDIS)]
        self.LAC_NM = ['LSPH', 'LSCO', 'LTCO', 'LAST', 'LPFC', 'LSFC', 'LTFC', 'LDIS']

        self.SAC_AN = 'Seidel Aberration Coefficents'
        self.SCW_AN = 'Seidel coefficients in waves'
        self.TAC_AN = 'Transverse Aberration Coefficents'
        self.LAC_AN = 'Longitudinal Aberration Coefficents'
        self.AberNames = [self.SAC_AN, self.SCW_AN, self.TAC_AN, self.LAC_AN]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep  9 07:54:16 2021

@author: joelherreravazquez & Antonio A.
"""

import numpy as np
import matplotlib.pyplot as plt
from .MathShapesClass import *



def psf(COEF, Focal, Diameter, Wave, pixels=600, PupilSample=4,  plot=0, sqr=0):

    ''' Generando los polinomios de Zernike (Nomenclarura Noll)con:
                   zernike_expand de Jherrera
    '''
    L=38 # Numero de terminos para la expanción polinomial


    ########################################################################

    '''Utilizando un solo valor de la pupila (x,y) con
                  Wavefront_Zernike_Phase
    '''
    Zern_pol, z_pow = zernike_expand(len(COEF))

    # ####################################################################

    """
            Generando un mapa de pupila aberrado
    """

    #Número de Elementos en la Matriz
    N=pixels
    #Variable de Muestreo
    Q=PupilSample
    #Diámetro de la Apertura [m]
    D=Diameter/1000.0
    #Diámetro de la Obstrucción[m]
    Do = 0.5
    #Número F
    FocalD=Focal/1000.0
    #Longitud de Onda [m]
    wvl=Wave*1e-6 # Cambiando wave a metros




    TamImag = int(N)
    r = (TamImag / (Q*2.0))

    center = int((TamImag / 2.0))
    xy=np.arange(0,TamImag)
    [X,Y]=np.meshgrid(xy, xy)
    x = ((X - center) / r)
    y = ((Y - center) / r)
    R=np.sqrt((x**2)+(y**2))


    W = Wavefront_Zernike_Phase(x, y, COEF)

    f=R>1
    W[f]=0.0

    T=np.copy(W)
    f=R<=1
    T[f]=1.0
    #Area del Círculo
    a=np.sum(np.sum(T))

    #Función de Pupila
    ##############################################################
    #Mapa del Frente de Onda con Máscara [wvl rms]
    # Wt=W*T
    #Campo complejo
    U=T*np.exp(-1j*2*np.pi*W)

    #PSF-Fraunhofer
    ##############################################################
    #TF de la distribución de Amplitud en Pupila
    F0=np.fft.fftshift(np.fft.fft2(np.fft.fftshift(U)))/a
    #Irradiancia en el Plano de Observación
    I=np.abs(F0)**2

    #Vector de Muestras con Cero en el Centro
    ##############################################################
    #Vector de Muestras
    v=np.arange(0,N)
    #Elemento central [bin]
    c=np.floor(N/2.)
    c=int(c)
    #Corriendo el cero al centro
    vx=v-c
    vy=c-v

    #Coordenadas del Plano de la Apertura
    ##############################################################
    #Largo del Plano
    L=Q*D
    ##############################################################
    #Coordenadas del Plano de Imagen
    #Tamaño del paso en frecuencia [m^-1]
    f=1/L
    #Vector de Coordenadas fx y fy
    fx=f*vx
    fy=f*vy

    #Vector de Coordenadas x y y
    u=fx*wvl*FocalD
    v=fy*wvl*FocalD


    umx=u[N-1]*1e6
    umn=u[0]*1e6
    vmx=v[N-1]*1e6
    vmn=v[0]*1e6

    I = np.rot90(I)
    #Saturando la Imagen




    if plot !=0:
        plt.figure(1)
        if sqr == 0:
            plt.imshow(I,extent=[umn,umx,vmx,vmn], cmap= plt.cm.bone)
            plt.colorbar()
            plt.ylabel('V[μm]')
            plt.xlabel('U[μm]')
            plt.title('Fraunhofer Prop - PSF')

        if sqr == 1:
            II=np.sqrt(I/np.max(I))*255
            plt.imshow(II,extent=[umn,umx,vmx,vmn], cmap= plt.cm.bone)
            plt.colorbar()
            plt.ylabel('V[μm]')
            plt.xlabel('U[μm]')
            plt.title('Fraunhofer Prop - PSF ( note: sqrt(I) )')

        if sqr == 2:
            III = np.log(I)
            III = III-np.min(III)
            III = 255*III/np.max(III)

            plt.imshow(III,extent=[umn,umx,vmx,vmn], cmap= plt.cm.bone)
            plt.colorbar()
            plt.ylabel('V[μm]')
            plt.xlabel('U[μm]')
            plt.title('Fraunhofer Prop - PSF ( note: sqrt(I) )')


    return I







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Solid Objects STL ARRAY"""

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos

# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Thickness = 5000.0
P_Obj.Glass = "AIR"
P_Obj.Diameter = 6.796727741707513E+002 * 2.0
P_Obj.Drawing = 0

# ______________________________________#

FOV = 0.5
Ref_esp = 0.8
Tf = 200  # 40
A1 = Tf * Tf
Conc = 800
Conc = Conc / 0.8
fpl = int(np.round(np.sqrt(Conc) / 2.0))
print("Numero de facetas por lado (Calculo 1)", (fpl * 2.0) + 1.0)
print("Tamanio por lado(mm) ", Tf * ((fpl * 2.0) + 1.0))

# ______________________________________#

n = fpl
FN = 1
focal = Tf * (fpl * 2.0 + 1.0) * FN
print("Distancia focal(mm) ", focal)

# ______________________________________#

sobredim = 2.0 * focal * np.tan(np.deg2rad(FOV / 2.0))
print("Sobredim (mm): ", sobredim)

# ______________________________________#

Tf = Tf - sobredim
A2 = Tf * Tf
RA = A1 / A2
Conc = Conc * RA
print("Nuevo tamaño de las facetas: ", Tf)

# ______________________________________#

fpl = int(np.round(np.sqrt(Conc) / 2.0))
print("Nuevo numero de facetas por lado (Calculo 2)", (fpl * 2.0) + 1.0)
print("Tamanio por lado 2a (mm) ", Tf * ((fpl * 2.0) + 1.0))

# ______________________________________#

n = fpl
Cx = Tf
Cy = Tf
Cz = 0
Lx = Tf
Ly = Tf
Lz = 1.0

# ______________________________________#

element0 = pv.Cube(center=(0.0, 0.0, 0.0), x_length=0.1, y_length=0.1, z_length=0.1, bounds=None)
for A in range(-n, n + 1):
    for B in range(-n, n + 1):
        Ty = 0.5 * np.rad2deg(np.arctan2(Cx * A, focal))
        Tx = -0.5 * np.rad2deg(np.arctan2(Cy * B, focal))
        element1 = pv.Cube(center=(0.0, 0.0, 0.0), x_length=Lx, y_length=Ly, z_length=Lz, bounds=None)
        element1.rotate_x(Tx)
        element1.rotate_y(Ty)
        v = [-Cx / 2.0, -Cy / 2.0, 0]
        element1.translate(v)
        v = [Cx * A, Cy * B, Cz]
        element1.translate(v)
        element0 = element0 + element1
element0.save("salida.stl")
direc = r"salida.stl"

# ______________________________________#

objeto = Kos.surf()
objeto.Diameter = 118.0 * 2.0
objeto.Solid_3d_stl = direc
objeto.Thickness = -6000
objeto.Glass = "MIRROR"
objeto.TiltX = 0
objeto.TiltY = 0
objeto.DespX = 0
objeto.DespY = 0
objeto.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Rc = 0
P_Ima.Thickness = -1.0
P_Ima.Glass = "AIR"
P_Ima.Diameter = 2000.0
P_Ima.Drawing = 1
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, objeto, P_Ima]
configur = Kos.Setup()

# ______________________________________#

Telescope = Kos.system(A, configur)
Rays = Kos.raykeeper(Telescope)

# ______________________________________#

W = 0.633
tam = 25
rad = 5500.0
tsis = len(A) + 2
for j in range(-tam, tam + 1):
    for i in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            Telescope.NsTrace(pSource_0, dCos, W)
            if np.shape(Telescope.NAME)[0] != 0:
                if Telescope.NAME[-1] == "Plano imagen":
                    plt.plot(Telescope.Hit_x[-1], Telescope.Hit_y[-1], '.', c="g")
                    Rays.push()

# ______________________________________#

plt.axis('square')
plt.show()
Kos.display3d(Telescope, Rays, 0)




import numpy as np
import scipy

def R_RMS_delta(delta_Z, L, M, N, X, Y):
    """R_RMS_delta.

    Parameters
    ----------
    delta_Z :
        delta_Z
    L :
        L
    M :
        M
    N :
        N
    X :
        X
    Y :
        Y
    """
    X = (((L / N) * delta_Z) + X)
    Y = (((M / N) * delta_Z) + Y)
    cenX = np.mean(X)
    cenY = np.mean(Y)
    x1 = (X - cenX)
    y1 = (Y - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return R_RMS

def RMS(X, Y, Z, L, M, N):
    """RMS.

    Parameters
    ----------
    X :
        X
    Y :
        Y
    Z :
        Z
    L :
        L
    M :
        M
    N :
        N
    """
    cenX = np.mean(X)
    cenY = np.mean(Y)
    x1 = (X - cenX)
    y1 = (Y - cenY)
    R2 = ((x1 * x1) + (y1 * y1))
    R_RMS = np.sqrt(np.mean(R2))
    return (R_RMS, cenX, cenY)

def BestFocus(X, Y, Z, L, M, N):
    """BestFocus.

    Parameters
    ----------
    X :
        X
    Y :
        Y
    Z :
        Z
    L :
        L
    M :
        M
    N :
        N
    """
    delta_Z = 0
    ZZ = (L, M, N, X, Y)
    v = scipy.optimize.fsolve(R_RMS_delta, delta_Z, args=ZZ)
    return v[0]




# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp Tel 2M Error Map"""

import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
import time


# ______________________________________#

def ErrorGen():
    L = 1000.
    N = 20.
    hight = 0.001
    SPACE = 2 * L / N
    x = np.arange(-L, L + SPACE, SPACE)
    y = np.arange(-L, L + SPACE, SPACE)
    gx, gy = np.meshgrid(x, y)
    R = np.sqrt((gx * gx) + (gy * gy))
    arg = np.argwhere(R < L)
    Npoints = np.shape(arg)[0]
    X = np.zeros(Npoints)
    Y = np.zeros(Npoints)
    i = 0
    for [a, b] in arg:
        X[i] = gx[a, b]
        Y[i] = gy[a, b]
        i = i + 1
    spa = 10000000
    Z = hight * (np.random.randint(-spa, spa, Npoints)) / (spa * 2.0)
    return [X, Y, Z, SPACE]


# ______________________________________#

P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 3500
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1.059E+003 * 2.0

# ______________________________________#

Thickness = 3.452200000000000E+003
M1 = Kos.surf()
M1.Rc = -9.638000000004009E+003
M1.Thickness = -Thickness
M1.k = -1.077310000000000E+000
M1.Glass = "MIRROR"
M1.Diameter = 1.059E+003 * 2.0
M1.InDiameter = 250 * 2.0
M1.Error_map = ErrorGen()

# ______________________________________#

M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness + 1.037525880125084E+003
M2.k = -4.328100000000000E+000
M2.Glass = "MIRROR"
M2.Diameter = 3.365E+002 * 2.0
M2.AxisMove = 0

# ______________________________________#

P_Ima = Kos.surf()
P_Ima.Diameter = 1000.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"

# ______________________________________#

A = [P_Obj, M1, M2, P_Ima]
configuracion_1 = Kos.Setup()

# ______________________________________#

Telescopio = Kos.system(A, configuracion_1)
Rayos = Kos.raykeeper(Telescopio)

# ______________________________________#

tam = 9
rad = 2100 / 2
tsis = len(A) - 1

# ______________________________________#

start_time = time.time()

# ______________________________________#

for i in range(-tam, tam + 1):
    for j in range(-tam, tam + 1):
        x_0 = (i / tam) * rad
        y_0 = (j / tam) * rad
        r = np.sqrt((x_0 * x_0) + (y_0 * y_0))
        if r < rad:
            print(i)
            tet = 0.0
            pSource_0 = [x_0, y_0, 0.0]
            dCos = [0.0, np.sin(np.deg2rad(tet)), np.cos(np.deg2rad(tet))]
            W = 0.4
            Telescopio.Trace(pSource_0, dCos, W)
            Rayos.push()

# ______________________________________#

print("--- %s seconds ---" % (time.time() - start_time))
Kos.display3d(Telescopio, Rayos, 2)
print(Telescopio.EFFL)
X, Y, Z, L, M, N = Rayos.pick(-1)

# ______________________________________#

plt.plot(X, Y, 'x')
plt.xlabel('X')
plt.ylabel('Y')
plt.title('Spot Diagram')
plt.axis('square')
plt.show()



#!/usr/bin/env Python3
"""
 Kraken-Optical-Simulator (KrakenOS)
 Python - Exact ray tracing library

Joel Herrera V., Carlos Guerrero P., Morgan Rhaí Najera Roa, Anais Sotelo B., Ilse Plauchu F.

• joel@astro.unam.mx

KrakenOS (Kraken - Optical Simulator) is a python library based in Numpy, Matplotlib, PyVTK and PyVista libraries, it provides a three-dimensional optical systems visualization and ray tracing. This tool has been programed on the object-oriented paradigm. KrakenOS focuses on performing sequential and non-sequential exact ray tracing, it permits to define all the parameters of the optical elements or even the mathematical function to describe their shape, it also allows adding optical properties to 3D solid elements in STL format and use glass catalogs. The library permit to control and modifying the position of the surfaces in a three-dimensional space, this allows generating off-axis systems. It also has several tools such as the calculation of wavefront aberrations in terms of Zernike polynomials, Seidel sums, Entrance and exit pupil calculation and paraxial optics.

PREREQUISITES
The library has been tested with the following packages and versions.
• Python '3.7.4'
• numpy '1.18.5'
• pyvista '0.25.3'
• pyvtk '0.5.18'
• matplotlib '3.4.3'
• vtk '8.2'
• csv '1.0'
• Place the directory “KrakenOS” in the same path where the code to be executed is located.

Surfaces and the optical system
The library has been simplified to the point of having only two classes of objects for the definition of a system, these are surf and •   system.
The surf object contains all the relevant information of every optical interface, in this way, every optical interface is an object of the surf class, all interfaces, from the object plane to the image plane, contain attributes of size, shape, material or orientation.

SURF CLASS ATRIBUTES:

  surf.Name = ""
     Name of the element.

  surf.NamePos = (0,0)
     “Name” position in the 2D diagram.

  surf.Note = "None"
     Useful for adding user notes to a surface.

  surf.Rc = 0
     Paraxial radius of curvature in millimeters.

  surf.Cylinder_Rxy_Ratio = 1
     Ratio between the axial and sagittal radius of curvature.

  surf.Axicon = 0
     values other than zero an axicon is generated with the angle defined

  surf.Thickness = 0.0
     Distance between this surface and the next surface.

  surf.Diameter = 1.0
     Outside diameter of the surface.

  surf.InDiameter = 0.0
     Internal diameter of the surface.

  surf.k = 0.0
     Conicity constant for classical conic surfaces, k = 0 for spherical, k = -1 for parabola, etc. Default value: 0.0

  surf.DespX = 0.0
  surf.DespY = 0.0
  surf.DespZ = 0.0
    Displacement of the surface in the X, Y and Z axis

  surf.TiltX = 0.0
  surf.TiltY = 0.0
  surf.TiltZ = 0.0
     Rotation of the surface in the X, Y and Z axis

  surf.Order = 0
     Define the order of the transformations.

  surf.AxisMove = 1
     Defines what will happen to the optical axis after a coordinate transformation.

  surf.Diff_Ord = 0.0
     Diffraction order.

  surf.Grating_D = 0.0
     Separation between the lines of the diffraction grating.

  surf.Grating_Angle = 0.0
     Angle of the grating lines in the plane of the surface

  surf.ZNK = np.zeros ()
     Zernike polynomials coefficients

  surf.ShiftX = 0
  surf.ShiftY = 0
     Offset the surface function on the X or Y axis.

  surf.Mask = 0
     (0) Do not apply mask, (1) Use mask as aperture, (2) Use mask as obstruction. Default value: 0

  surf.Mask_Shape = Object_3D
     Form of the mask to apply on surface

  surf.AspherData = np.zeros ()
     Asphericity coefficients.

  surf.ExtraData = [f, coef]
     User-defined function for optical surface

  surf.Error_map = [X, Y, Z, SPACE]
     Error map array

  surf.Drawing = 1
     1 for drawn in the 3D plot, 0 to omit.

  surf.Color = [0,0,0]
     Element color for 3D Plot. [R,G,B]

  surf.Solid_3d_stl = "None"
     Path to the 3D solid STL file.


SYSTEM CLASS ATRIBUTES:

  system.Trace (pS, dC, wV)
     Sequential ray tracing.
     pS = [1.0, 0.0, 0.0] – Ray origin coordinates
     dC = [0.0,0.0,1.0] - The directing cosines
     wV = 0.4 - Wavelength

  system.NsTrace (pS, dC, wV)
     Non-Sequential ray tracing

  system.Parax (w)
     Paraxial optics calculations

  system.disable_inner
     Enables the central aperture.

  system.enable_inner
	Disables the central aperture.

  system.SURFACE
	Returns the surfaces the ray passed through.

  system.NAME
	Returns surface names that the ray passed through

  system.GLASS
	Returns materials that the ray passed through.

  system.XYZ
	[X, Y, Z] ray coordinates from its origin to the image plane.

  system.OST_XYZ
	[X, Y, Z] coordinates of ray intersections with respect to a coordinate system at its vertex, even if this vertex has a translation or rotation.

  system.DISTANCE
     List of distances traveled by the ray.

  system.OP
	List of optical paths.

  system.TOP
	Total optical path.

  system.TOP_S     List of the ray's optical path by sections.

  system.ALPHA
     List the materials absorption coefficients

  system.BULK_TRANS
     List the transmission through all the •   system. absorption coefficients are considered.

  system.S_LMN
     Surface normal direction cosines [L, M, N].

  system.LMN
     incident ray direction cosines [L, M, N].

  system.R_LMN
     Resulting ray direction cosines [L, M, N].

  system.N0
     Refractive indices before and after each interface

  system.N1
     Refractive indices after each interface. This is useful to differentiate between index before and after an iteration. Example:
     N0 = [n1, n2, n3, n4, n5]  N1 = [n2, n3, n4, n5, n5]

  system.WAV
	Wavelength of the ray (µm)

  system.G_LMN
 	[L, M, N] Direction cosines that define the lines of the diffraction grating on the plane.

  system.ORDER
	      Ray diffraction order.

  system.GRATING_D
	 Distance between lines of the diffraction grating.Units (µm)

  system.RP
	Fresnel reflection coefficients for polarization P.

  system.RS
	Fresnel reflection coefficients for polarization S.

  system.TP
	Fresnel transmission coefficients for polarization P.

  system.TS
	Fresnel transmission coefficients for polarization S.

  system.TTBE
	Total energy transmitted or reflected per element.

  system.TT
	Total energy transmitted or reflected total.

  system.targ_surf (int)
    Limits the ray tracing to the defined surface

  system.flat_surf (int)
    Change a surface to flat.



USER MANUAL AND EXAMPLES
Very important, read the included user manual (KrakenOS_User_Manual.pdf) and the set of useful examples:

• Examp_Axicon.py
• Examp_Axicon_And_Cylinder.py
• Examp_Diffraction_Grating_Reflection.py
• Examp_Diffraction_Grating_Transmission.py
• Examp_Doublet_Lens_ParaxMatrix.py
• Examp_Doublet_Lens.py
• Examp_Doublet_Lens_3Dcolor.py
• Examp_Doublet_Lens_Commands•   system.py
• Examp_Doublet_Lens_Cylinder.py
• Examp_Doublet_Lens_NonSec.py
• Examp_Doublet_Lens_Pupil.py
• Examp_Doublet_Lens_Pupil_Seidel.py
• Examp_Doublet_Lens_Tilt_Nulls.py
• Examp_Doublet_Lens_Tilt.py
• Examp_Doublet_Lens_Tilt_non_sec.py
• Examp_Doublet_Lens_Zernike.py
• Examp_ExtraShape_Micro_Lens_Array.py
• Examp_ExtraShape_Radial_Sine.py
• Examp_ExtraShape_XY_Cosines.py
• Examp_Flat_Mirror_45Deg.py
• Examp_MultiCore.py
• Examp_ParaboleMirror_Shift.py
• Examp_Perfect_lens.py
• Examp_Ray.py
• Examp_Solid_Object_STL.py
• Examp_Solid_Object_STL_ARRAY.py
• Examp_Sourse_Distribution_Function.py
• Examp_Tel_2M.py
• Examp_Tel_2M_Error_Map.py
• Examp_Tel_2M_Pupila.py
• Examp_Tel_2M_Spyder_Spot_Diagram.py
• Examp_Tel_2M_Spyder_Spot_RMS.py
• Examp_Tel_2M_Spyder_Spot_Tilt_M2.py
• Examp_Tel_2M_Atmospheric_Refraction.py
• Examp_Tel_2M_Wavefront_Fitting.py

Enjoy it!!

"""


# import inspect

from .Physics import *
from .Display import *
from .PhysicsClass import *
from .SurfClass import *
from .PupilTool import *
from .RayKeeper import *
from .SystemTools import *
from .SetupClass import *
from .SurfTools import surface_tools as SUT
from .Prerequisites3D import *
from .HitOnSurf import *
from .InterNormalCalc import *
from .WavefrontFit import *
from .SeidelTool import *
from .SourceRand import *
from .LibRMS import *
from .TraceLoopTool import *
from .PhaseCalc import *
from .WavePlot import *
from .KrakenSys import *
from .PSFCalc import *




# RUTE=inspect.getmodule(SUT)
# print("-------------- rute ------------")
# print(RUTE)


# RUTE=inspect.getmodule(SUT).__file__
# RUTE=RUTE[:-12]

# print (RUTE)



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examp-Tel_2M_Atmospheric_Refraction_Corrector"""

import pkg_resources
required = {'KrakenOS'}
installed = {pkg.key for pkg in pkg_resources.working_set}
missing = required - installed

if missing:
    print("No instalado")
    import sys
    sys.path.append("../..")


import KrakenOS as Kos
import numpy as np
import matplotlib.pyplot as plt
import scipy

# _________________________________________________________________#
P_Obj = Kos.surf()
P_Obj.Rc = 0
P_Obj.Thickness = 1000 + 3452.2
P_Obj.Glass = "AIR"
P_Obj.Diameter = 1059.0 * 2.0
P_Obj.Drawing = 0
# _________________________________________________________________#
Thickness = 3452.2
M1 = Kos.surf()
M1.Rc = -9638.0
M1.Thickness = -Thickness
M1.k = -1.07731
M1.Glass = "MIRROR"
M1.Diameter = 1059.0 * 2.0
M1.InDiameter = 250 * 2.0
# _________________________________________________________________#
M2 = Kos.surf()
M2.Rc = -3.93E+003
M2.Thickness = Thickness + 1037.525 - 300.0
M2.k = -4.3281
M2.Glass = "MIRROR"
M2.Diameter = 336.5 * 2.0
M2.AxisMove = 0
# _________________________________________________________________#
C1 = Kos.surf()
C1.Thickness = 5
C1.Glass = "BK7"
C1.Diameter = 100

C2 = Kos.surf()
C2.Thickness = C1.Thickness
C2.Glass = "F2"
C2.Diameter = 100
C2.TiltY = 1.55
C2.AxisMove = 0

C3 = Kos.surf()
C3.Thickness = C1.Thickness + 288.631
C3.Glass = "AIR"
C3.Diameter = 100
C3.TiltY = 0
C3.AxisMove = 0

# -------------------------------------

C6 = Kos.surf()
C6.Thickness = C1.Thickness
C6.Glass = "AIR"
C6.Diameter = 100

# _________________________________________________________________#
P_Ima = Kos.surf()
P_Ima.Diameter = 100.0
P_Ima.Glass = "AIR"
P_Ima.Name = "Plano imagen"
A = [P_Obj, M1, M2, C1, C2, C3, P_Ima]
# _________________________________________________________________#
configuracion_1 = Kos.Setup()
Telescopio = Kos.system(A, configuracion_1)

Rayos1 = Kos.raykeeper(Telescopio)
Rayos2 = Kos.raykeeper(Telescopio)
Rayos3 = Kos.raykeeper(Telescopio)
Rayos = Kos.raykeeper(Telescopio)

W = 0.60169
sup = 1  # Difining M1 as enter pupil diameter
AperVal = 2000
AperType = "EPD"  # "STOP"
Pup = Kos.PupilCalc(Telescopio, sup, W, AperType, AperVal)
Pup.Samp = 5
Pup.FieldType = "angle"

Pup.AtmosRef = 1
Pup.T = 283.15  # k
Pup.P = 101300  # Pa
Pup.H = 0.5  # Humidity ratio 1 to 0
Pup.xc = 400  # ppm
Pup.lat = 31  # degrees
Pup.h = 2800  # meters
Pup.l1 = W  # micron
Pup.l2 = 0.50169  # micron
Pup.z0 = 55.0  # degrees

Pup.Ptype = "hexapolar"
Pup.FieldX = 0.0

W1 = 0.50169
Pup.l2 = W1
xa, ya, za, La, Ma, Na = Pup.Pattern2Field()

W2 = 0.60169
Pup.l2 = W2
xb, yb, zb, Lb, Mb, Nb = Pup.Pattern2Field()

W3 = 0.70169
Pup.l2 = W3
xc, yc, zc, Lc, Mc, Nc = Pup.Pattern2Field()

############################################


for i in range(0, len(xc)):
    pSource_0 = [xc[i], yc[i], zc[i]]
    dCos = [Lc[i], Mc[i], Nc[i]]
    Telescopio.Trace(pSource_0, dCos, W3)
    Rayos3.push()
    Rayos.push()

for i in range(0, len(xb)):
    pSource_0 = [xb[i], yb[i], zb[i]]
    dCos = [Lb[i], Mb[i], Nb[i]]
    Telescopio.Trace(pSource_0, dCos, W2)
    Rayos2.push()
    Rayos.push()

for i in range(0, len(xa)):
    pSource_0 = [xa[i], ya[i], za[i]]
    dCos = [La[i], Ma[i], Na[i]]
    Telescopio.Trace(pSource_0, dCos, W1)
    Rayos1.push()
    Rayos.push()

############################################

Kos.display3d(Telescopio, Rayos, 1)

X, Y, Z, L, M, N = Rayos1.pick(-1)
plt.plot(X * 1000.0, Y * 1000.0, 'x', c="b", ms=2)

X, Y, Z, L, M, N = Rayos3.pick(-1)
plt.plot(X * 1000.0, Y * 1000.0, 'x', c="g", ms=2)

X, Y, Z, L, M, N = Rayos2.pick(-1)
plt.plot(X * 1000.0, Y * 1000.0, 'x', c="r", ms=2)

# axis labeling
plt.xlabel('x')
plt.ylabel('y')

# figure name
plt.title('Spot Diagram')
plt.axis('square')
plt.ylim(-5, 5)
plt.show()

# v = Kos.BestFocus(X,Y,Z,L,M,N)
# rms = Kos.RMS(X,Y,Z,L,M,N)

# print("sol ---------------------")
# print(v)
